# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


class CreateEnvironment(AAZCommand):
    """Create virtual machines in a lab. This operation can take a while to complete.

    :example: Create a VM in the lab from a gallery image.
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Ubuntu Server 22.04 LTS" --image-type gallery --size Standard_DS1_v2

    :example: Create a VM in the lab from a gallery image with SSH authentication.
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Ubuntu Server 22.04 LTS" --image-type gallery --size Standard_DS1_v2 --authentication-type ssh

    :example: Create a claimable VM in the lab from a gallery image with password authentication.
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Ubuntu Server 22.04 LTS" --image-type gallery --size Standard_DS1_v2 --allow-claim

    :example: Create a windows VM in the lab from a gallery image with password authentication.
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Windows Server 2022 Datacenter" --image-type gallery --size Standard_DS1_v2

    :example: Create a VM in the lab from a custom image.
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "jenkins_custom" --image-type custom --size Standard_DS1_v2

    :example: Create a VM in the lab with a public IP.
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --image "Ubuntu Server 22.04 LTS" --image-type gallery --size Standard_DS1_v2 --ip-configuration public

    :example: Create a VM from a formula.
        az lab vm create --lab-name {LabName} -g {ResourceGroup} --name {VMName} --formula MyFormula --artifacts 'artifacts.json'
    """

    _aaz_info = {
        "version": "2018-09-15",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.devtestlab/labs/{}/createenvironment", "2018-09-15"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, None)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.lab_name = AAZStrArg(
            options=["--lab-name"],
            help="The name of the lab.",
            required=True,
            id_part="name",
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        _args_schema.location = AAZStrArg(
            options=["--location"],
            help="The location of the new virtual machine or environment",
        )
        _args_schema.name = AAZStrArg(
            options=["-n", "--name"],
            help="The name of the virtual machine or environment",
        )
        _args_schema.allow_claim = AAZBoolArg(
            options=["--allow-claim"],
            help="Flag indicating if the VM should be created as claimable.",
            default=False,
        )
        _args_schema.artifacts_org = AAZListArg(
            options=["--artifacts-org"],
            help="The artifacts to be installed on the virtual machine.",
        )
        _args_schema.bulk_creation_parameters = AAZObjectArg(
            options=["--bulk-creation-parameters"],
            help="The number of virtual machine instances to create.",
        )
        _args_schema.created_date = AAZDateTimeArg(
            options=["--created-date"],
            help="The creation date of the virtual machine.",
        )
        _args_schema.custom_image_id = AAZStrArg(
            options=["--custom-image-id"],
            help="The custom image identifier of the virtual machine.",
        )
        _args_schema.data_disk_parameters = AAZListArg(
            options=["--data-disk-parameters"],
            help="New or existing data disks to attach to the virtual machine after creation",
        )
        _args_schema.disallow_public_ip_address = AAZBoolArg(
            options=["--disallow-public-ip-address"],
            help="Indicates whether the virtual machine is to be created without a public IP address.",
            default=False,
        )
        _args_schema.environment_id = AAZStrArg(
            options=["--environment-id"],
            help="The resource ID of the environment that contains this virtual machine, if any.",
        )
        _args_schema.expiration_date = AAZDateTimeArg(
            options=["--expiration-date"],
            help="The expiration date for VM.",
        )
        _args_schema.gallery_image_reference = AAZObjectArg(
            options=["--gallery-image-reference"],
            help="The Microsoft Azure Marketplace image reference of the virtual machine.",
        )
        _args_schema.is_authentication_with_ssh_key = AAZBoolArg(
            options=["--is-authentication-with-ssh-key"],
            help="Indicates whether this virtual machine uses an SSH key for authentication.",
        )
        _args_schema.lab_subnet_name = AAZStrArg(
            options=["--lab-subnet-name"],
            help="The lab subnet name of the virtual machine.",
        )
        _args_schema.lab_virtual_network_id = AAZStrArg(
            options=["--lab-virtual-network-id"],
            help="The lab virtual network identifier of the virtual machine.",
        )
        _args_schema.network_interface = AAZObjectArg(
            options=["--network-interface"],
            help="The network interface properties.",
        )
        _args_schema.notes = AAZStrArg(
            options=["--notes"],
            help="The notes of the virtual machine.",
        )
        _args_schema.owner_object_id = AAZStrArg(
            options=["--owner-object-id"],
            help="The object identifier of the owner of the virtual machine.",
            default="dynamicValue",
        )
        _args_schema.owner_user_principal_name = AAZStrArg(
            options=["--owner-user-principal-name"],
            help="The user principal name of the virtual machine owner.",
        )
        _args_schema.password = AAZStrArg(
            options=["--password"],
            help="The password of the virtual machine administrator.",
        )
        _args_schema.plan_id = AAZStrArg(
            options=["--plan-id"],
            help="The id of the plan associated with the virtual machine image",
        )
        _args_schema.schedule_parameters = AAZListArg(
            options=["--schedule-parameters"],
            help="Virtual Machine schedules to be created",
        )
        _args_schema.size = AAZStrArg(
            options=["--size"],
            help="The size of the VM to be created. See                                      https://azure.microsoft.com/pricing/details/virtual-machines/ for size info.",
        )
        _args_schema.ssh_key = AAZStrArg(
            options=["--ssh-key"],
            help="The SSH public key or public key file path. Use `--generate-ssh-keys` to generate SSH keys.",
        )
        _args_schema.storage_type = AAZStrArg(
            options=["--storage-type"],
            help="Storage type to use for virtual machine (i.e. Standard, Premium).",
            default="labStorageType",
        )
        _args_schema.user_name = AAZStrArg(
            options=["--user-name"],
            help="The user name of the virtual machine.",
        )
        _args_schema.tags = AAZDictArg(
            options=["--tags"],
            help="The tags of the resource.",
        )

        artifacts_org = cls._args_schema.artifacts_org
        artifacts_org.Element = AAZObjectArg()

        _element = cls._args_schema.artifacts_org.Element
        _element.artifact_id = AAZStrArg(
            options=["artifact-id"],
            help="The artifact's identifier.",
        )
        _element.artifact_title = AAZStrArg(
            options=["artifact-title"],
            help="The artifact's title.",
        )
        _element.deployment_status_message = AAZStrArg(
            options=["deployment-status-message"],
            help="The status message from the deployment.",
        )
        _element.install_time = AAZDateTimeArg(
            options=["install-time"],
            help="The time that the artifact starts to install on the virtual machine.",
        )
        _element.parameters = AAZListArg(
            options=["parameters"],
            help="The parameters of the artifact.",
        )
        _element.status = AAZStrArg(
            options=["status"],
            help="The status of the artifact.",
        )
        _element.vm_extension_status_message = AAZStrArg(
            options=["vm-extension-status-message"],
            help="The status message from the virtual machine extension.",
        )

        parameters = cls._args_schema.artifacts_org.Element.parameters
        parameters.Element = AAZObjectArg()

        _element = cls._args_schema.artifacts_org.Element.parameters.Element
        _element.name = AAZStrArg(
            options=["name"],
            help="The name of the artifact parameter.",
        )
        _element.value = AAZStrArg(
            options=["value"],
            help="The value of the artifact parameter.",
        )

        bulk_creation_parameters = cls._args_schema.bulk_creation_parameters
        bulk_creation_parameters.instance_count = AAZIntArg(
            options=["instance-count"],
            help="The number of virtual machine instances to create.",
        )

        data_disk_parameters = cls._args_schema.data_disk_parameters
        data_disk_parameters.Element = AAZObjectArg()

        _element = cls._args_schema.data_disk_parameters.Element
        _element.attach_new_data_disk_options = AAZObjectArg(
            options=["attach-new-data-disk-options"],
            help="Specifies options to attach a new disk to the virtual machine.",
        )
        _element.existing_lab_disk_id = AAZStrArg(
            options=["existing-lab-disk-id"],
            help="Specifies the existing lab disk id to attach to virtual machine.",
        )
        _element.host_caching = AAZStrArg(
            options=["host-caching"],
            help="Caching option for a data disk (i.e. None, ReadOnly, ReadWrite).",
            enum={"None": "None", "ReadOnly": "ReadOnly", "ReadWrite": "ReadWrite"},
        )

        attach_new_data_disk_options = cls._args_schema.data_disk_parameters.Element.attach_new_data_disk_options
        attach_new_data_disk_options.disk_name = AAZStrArg(
            options=["disk-name"],
            help="The name of the disk to be attached.",
        )
        attach_new_data_disk_options.disk_size_gi_b = AAZIntArg(
            options=["disk-size-gi-b"],
            help="Size of the disk to be attached in Gibibytes.",
        )
        attach_new_data_disk_options.disk_type = AAZStrArg(
            options=["disk-type"],
            help="The storage type for the disk (i.e. Standard, Premium).",
            enum={"Premium": "Premium", "Standard": "Standard", "StandardSSD": "StandardSSD"},
        )

        gallery_image_reference = cls._args_schema.gallery_image_reference
        gallery_image_reference.offer = AAZStrArg(
            options=["offer"],
            help="The offer of the gallery image.",
        )
        gallery_image_reference.os_type = AAZStrArg(
            options=["os-type"],
            help="The OS type of the gallery image.",
        )
        gallery_image_reference.publisher = AAZStrArg(
            options=["publisher"],
            help="The publisher of the gallery image.",
        )
        gallery_image_reference.sku = AAZStrArg(
            options=["sku"],
            help="The SKU of the gallery image.",
        )
        gallery_image_reference.version = AAZStrArg(
            options=["version"],
            help="The version of the gallery image.",
        )

        network_interface = cls._args_schema.network_interface
        network_interface.dns_name = AAZStrArg(
            options=["dns-name"],
            help="The DNS name.",
        )
        network_interface.private_ip_address = AAZStrArg(
            options=["private-ip-address"],
            help="The private IP address.",
        )
        network_interface.public_ip_address = AAZStrArg(
            options=["public-ip-address"],
            help="The public IP address.",
        )
        network_interface.public_ip_address_id = AAZStrArg(
            options=["public-ip-address-id"],
            help="The resource ID of the public IP address.",
        )
        network_interface.rdp_authority = AAZStrArg(
            options=["rdp-authority"],
            help="The RdpAuthority property is a server DNS host name or IP address followed by the service port number for RDP (Remote Desktop Protocol).",
        )
        network_interface.shared_public_ip_address_configuration = AAZObjectArg(
            options=["shared-public-ip-address-configuration"],
            help="The configuration for sharing a public IP address across multiple virtual machines.",
        )
        network_interface.ssh_authority = AAZStrArg(
            options=["ssh-authority"],
            help="The SshAuthority property is a server DNS host name or IP address followed by the service port number for SSH.",
        )
        network_interface.subnet_id = AAZStrArg(
            options=["subnet-id"],
            help="The resource ID of the sub net.",
        )
        network_interface.virtual_network_id = AAZStrArg(
            options=["virtual-network-id"],
            help="The resource ID of the virtual network.",
        )

        shared_public_ip_address_configuration = cls._args_schema.network_interface.shared_public_ip_address_configuration
        shared_public_ip_address_configuration.inbound_nat_rules = AAZListArg(
            options=["inbound-nat-rules"],
            help="The incoming NAT rules",
        )

        inbound_nat_rules = cls._args_schema.network_interface.shared_public_ip_address_configuration.inbound_nat_rules
        inbound_nat_rules.Element = AAZObjectArg()

        _element = cls._args_schema.network_interface.shared_public_ip_address_configuration.inbound_nat_rules.Element
        _element.backend_port = AAZIntArg(
            options=["backend-port"],
            help="The port to which the external traffic will be redirected.",
        )
        _element.frontend_port = AAZIntArg(
            options=["frontend-port"],
            help="The external endpoint port of the inbound connection. Possible values range between 1 and 65535, inclusive. If unspecified, a value will be allocated automatically.",
        )
        _element.transport_protocol = AAZStrArg(
            options=["transport-protocol"],
            help="The transport protocol for the endpoint.",
            enum={"Tcp": "Tcp", "Udp": "Udp"},
        )

        schedule_parameters = cls._args_schema.schedule_parameters
        schedule_parameters.Element = AAZObjectArg()

        _element = cls._args_schema.schedule_parameters.Element
        _element.name = AAZStrArg(
            options=["name"],
            help="The name of the virtual machine or environment",
        )
        _element.daily_recurrence = AAZObjectArg(
            options=["daily-recurrence"],
            help="If the schedule will occur once each day of the week, specify the daily recurrence.",
        )
        _element.hourly_recurrence = AAZObjectArg(
            options=["hourly-recurrence"],
            help="If the schedule will occur multiple times a day, specify the hourly recurrence.",
        )
        _element.notification_settings = AAZObjectArg(
            options=["notification-settings"],
            help="Notification settings.",
        )
        _element.status = AAZStrArg(
            options=["status"],
            help="The status of the schedule (i.e. Enabled, Disabled)",
            default="Disabled",
            enum={"Disabled": "Disabled", "Enabled": "Enabled"},
        )
        _element.target_resource_id = AAZStrArg(
            options=["target-resource-id"],
            help="The resource ID to which the schedule belongs",
        )
        _element.task_type = AAZStrArg(
            options=["task-type"],
            help="The task type of the schedule (e.g. LabVmsShutdownTask, LabVmAutoStart).",
        )
        _element.time_zone_id = AAZStrArg(
            options=["time-zone-id"],
            help="The time zone ID (e.g. China Standard Time, Greenland Standard Time, Pacific Standard time, etc.). The possible values for this property can be found in `IReadOnlyCollection<string> TimeZoneConverter.TZConvert.KnownWindowsTimeZoneIds` (https://github.com/mattjohnsonpint/TimeZoneConverter/blob/main/README.md)",
        )
        _element.weekly_recurrence = AAZObjectArg(
            options=["weekly-recurrence"],
            help="If the schedule will occur only some days of the week, specify the weekly recurrence.",
        )
        _element.tags = AAZDictArg(
            options=["tags"],
            help="The tags of the resource.",
        )

        daily_recurrence = cls._args_schema.schedule_parameters.Element.daily_recurrence
        daily_recurrence.time = AAZStrArg(
            options=["time"],
            help="The time of day the schedule will occur.",
        )

        hourly_recurrence = cls._args_schema.schedule_parameters.Element.hourly_recurrence
        hourly_recurrence.minute = AAZIntArg(
            options=["minute"],
            help="Minutes of the hour the schedule will run.",
        )

        notification_settings = cls._args_schema.schedule_parameters.Element.notification_settings
        notification_settings.email_recipient = AAZStrArg(
            options=["email-recipient"],
            help="The email recipient to send notifications to (can be a list of semi-colon separated email addresses).",
        )
        notification_settings.notification_locale = AAZStrArg(
            options=["notification-locale"],
            help="The locale to use when sending a notification (fallback for unsupported languages is EN).",
        )
        notification_settings.status = AAZStrArg(
            options=["status"],
            help="If notifications are enabled for this schedule (i.e. Enabled, Disabled).",
            default="Disabled",
            enum={"Disabled": "Disabled", "Enabled": "Enabled"},
        )
        notification_settings.time_in_minutes = AAZIntArg(
            options=["time-in-minutes"],
            help="Time in minutes before event at which notification will be sent.",
        )
        notification_settings.webhook_url = AAZStrArg(
            options=["webhook-url"],
            help="The webhook URL to which the notification will be sent.",
        )

        weekly_recurrence = cls._args_schema.schedule_parameters.Element.weekly_recurrence
        weekly_recurrence.time = AAZStrArg(
            options=["time"],
            help="The time of the day the schedule will occur.",
        )
        weekly_recurrence.weekdays = AAZListArg(
            options=["weekdays"],
            help="The days of the week for which the schedule is set (e.g. Sunday, Monday, Tuesday, etc.).",
        )

        weekdays = cls._args_schema.schedule_parameters.Element.weekly_recurrence.weekdays
        weekdays.Element = AAZStrArg()

        tags = cls._args_schema.schedule_parameters.Element.tags
        tags.Element = AAZStrArg()

        tags = cls._args_schema.tags
        tags.Element = AAZStrArg()
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.LabsCreateEnvironment(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    class LabsCreateEnvironment(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.DevTestLab/labs/{lab-name}/createEnvironment",
                **self.url_parameters
            )

        @property
        def method(self):
            return "POST"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "lab-name", self.ctx.args.lab_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2018-09-15",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("location", AAZStrType, ".location")
            _builder.set_prop("name", AAZStrType, ".name")
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})
            _builder.set_prop("tags", AAZDictType, ".tags")

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("allowClaim", AAZBoolType, ".allow_claim")
                properties.set_prop("artifacts", AAZListType, ".artifacts_org")
                properties.set_prop("bulkCreationParameters", AAZObjectType, ".bulk_creation_parameters")
                properties.set_prop("createdDate", AAZStrType, ".created_date")
                properties.set_prop("customImageId", AAZStrType, ".custom_image_id")
                properties.set_prop("dataDiskParameters", AAZListType, ".data_disk_parameters")
                properties.set_prop("disallowPublicIpAddress", AAZBoolType, ".disallow_public_ip_address")
                properties.set_prop("environmentId", AAZStrType, ".environment_id")
                properties.set_prop("expirationDate", AAZStrType, ".expiration_date")
                properties.set_prop("galleryImageReference", AAZObjectType, ".gallery_image_reference")
                properties.set_prop("isAuthenticationWithSshKey", AAZBoolType, ".is_authentication_with_ssh_key")
                properties.set_prop("labSubnetName", AAZStrType, ".lab_subnet_name")
                properties.set_prop("labVirtualNetworkId", AAZStrType, ".lab_virtual_network_id")
                properties.set_prop("networkInterface", AAZObjectType, ".network_interface")
                properties.set_prop("notes", AAZStrType, ".notes")
                properties.set_prop("ownerObjectId", AAZStrType, ".owner_object_id")
                properties.set_prop("ownerUserPrincipalName", AAZStrType, ".owner_user_principal_name")
                properties.set_prop("password", AAZStrType, ".password", typ_kwargs={"flags": {"secret": True}})
                properties.set_prop("planId", AAZStrType, ".plan_id")
                properties.set_prop("scheduleParameters", AAZListType, ".schedule_parameters")
                properties.set_prop("size", AAZStrType, ".size")
                properties.set_prop("sshKey", AAZStrType, ".ssh_key", typ_kwargs={"flags": {"secret": True}})
                properties.set_prop("storageType", AAZStrType, ".storage_type")
                properties.set_prop("userName", AAZStrType, ".user_name")

            artifacts = _builder.get(".properties.artifacts")
            if artifacts is not None:
                artifacts.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.artifacts[]")
            if _elements is not None:
                _elements.set_prop("artifactId", AAZStrType, ".artifact_id")
                _elements.set_prop("artifactTitle", AAZStrType, ".artifact_title")
                _elements.set_prop("deploymentStatusMessage", AAZStrType, ".deployment_status_message")
                _elements.set_prop("installTime", AAZStrType, ".install_time")
                _elements.set_prop("parameters", AAZListType, ".parameters")
                _elements.set_prop("status", AAZStrType, ".status")
                _elements.set_prop("vmExtensionStatusMessage", AAZStrType, ".vm_extension_status_message")

            parameters = _builder.get(".properties.artifacts[].parameters")
            if parameters is not None:
                parameters.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.artifacts[].parameters[]")
            if _elements is not None:
                _elements.set_prop("name", AAZStrType, ".name")
                _elements.set_prop("value", AAZStrType, ".value")

            bulk_creation_parameters = _builder.get(".properties.bulkCreationParameters")
            if bulk_creation_parameters is not None:
                bulk_creation_parameters.set_prop("instanceCount", AAZIntType, ".instance_count")

            data_disk_parameters = _builder.get(".properties.dataDiskParameters")
            if data_disk_parameters is not None:
                data_disk_parameters.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.dataDiskParameters[]")
            if _elements is not None:
                _elements.set_prop("attachNewDataDiskOptions", AAZObjectType, ".attach_new_data_disk_options")
                _elements.set_prop("existingLabDiskId", AAZStrType, ".existing_lab_disk_id")
                _elements.set_prop("hostCaching", AAZStrType, ".host_caching")

            attach_new_data_disk_options = _builder.get(".properties.dataDiskParameters[].attachNewDataDiskOptions")
            if attach_new_data_disk_options is not None:
                attach_new_data_disk_options.set_prop("diskName", AAZStrType, ".disk_name")
                attach_new_data_disk_options.set_prop("diskSizeGiB", AAZIntType, ".disk_size_gi_b")
                attach_new_data_disk_options.set_prop("diskType", AAZStrType, ".disk_type")

            gallery_image_reference = _builder.get(".properties.galleryImageReference")
            if gallery_image_reference is not None:
                gallery_image_reference.set_prop("offer", AAZStrType, ".offer")
                gallery_image_reference.set_prop("osType", AAZStrType, ".os_type")
                gallery_image_reference.set_prop("publisher", AAZStrType, ".publisher")
                gallery_image_reference.set_prop("sku", AAZStrType, ".sku")
                gallery_image_reference.set_prop("version", AAZStrType, ".version")

            network_interface = _builder.get(".properties.networkInterface")
            if network_interface is not None:
                network_interface.set_prop("dnsName", AAZStrType, ".dns_name")
                network_interface.set_prop("privateIpAddress", AAZStrType, ".private_ip_address")
                network_interface.set_prop("publicIpAddress", AAZStrType, ".public_ip_address")
                network_interface.set_prop("publicIpAddressId", AAZStrType, ".public_ip_address_id")
                network_interface.set_prop("rdpAuthority", AAZStrType, ".rdp_authority")
                network_interface.set_prop("sharedPublicIpAddressConfiguration", AAZObjectType, ".shared_public_ip_address_configuration")
                network_interface.set_prop("sshAuthority", AAZStrType, ".ssh_authority")
                network_interface.set_prop("subnetId", AAZStrType, ".subnet_id")
                network_interface.set_prop("virtualNetworkId", AAZStrType, ".virtual_network_id")

            shared_public_ip_address_configuration = _builder.get(".properties.networkInterface.sharedPublicIpAddressConfiguration")
            if shared_public_ip_address_configuration is not None:
                shared_public_ip_address_configuration.set_prop("inboundNatRules", AAZListType, ".inbound_nat_rules")

            inbound_nat_rules = _builder.get(".properties.networkInterface.sharedPublicIpAddressConfiguration.inboundNatRules")
            if inbound_nat_rules is not None:
                inbound_nat_rules.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.networkInterface.sharedPublicIpAddressConfiguration.inboundNatRules[]")
            if _elements is not None:
                _elements.set_prop("backendPort", AAZIntType, ".backend_port")
                _elements.set_prop("frontendPort", AAZIntType, ".frontend_port")
                _elements.set_prop("transportProtocol", AAZStrType, ".transport_protocol")

            schedule_parameters = _builder.get(".properties.scheduleParameters")
            if schedule_parameters is not None:
                schedule_parameters.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.scheduleParameters[]")
            if _elements is not None:
                _elements.set_prop("name", AAZStrType, ".name")
                _elements.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})
                _elements.set_prop("tags", AAZDictType, ".tags")

            properties = _builder.get(".properties.scheduleParameters[].properties")
            if properties is not None:
                properties.set_prop("dailyRecurrence", AAZObjectType, ".daily_recurrence")
                properties.set_prop("hourlyRecurrence", AAZObjectType, ".hourly_recurrence")
                properties.set_prop("notificationSettings", AAZObjectType, ".notification_settings")
                properties.set_prop("status", AAZStrType, ".status")
                properties.set_prop("targetResourceId", AAZStrType, ".target_resource_id")
                properties.set_prop("taskType", AAZStrType, ".task_type")
                properties.set_prop("timeZoneId", AAZStrType, ".time_zone_id")
                properties.set_prop("weeklyRecurrence", AAZObjectType, ".weekly_recurrence")

            daily_recurrence = _builder.get(".properties.scheduleParameters[].properties.dailyRecurrence")
            if daily_recurrence is not None:
                daily_recurrence.set_prop("time", AAZStrType, ".time")

            hourly_recurrence = _builder.get(".properties.scheduleParameters[].properties.hourlyRecurrence")
            if hourly_recurrence is not None:
                hourly_recurrence.set_prop("minute", AAZIntType, ".minute")

            notification_settings = _builder.get(".properties.scheduleParameters[].properties.notificationSettings")
            if notification_settings is not None:
                notification_settings.set_prop("emailRecipient", AAZStrType, ".email_recipient")
                notification_settings.set_prop("notificationLocale", AAZStrType, ".notification_locale")
                notification_settings.set_prop("status", AAZStrType, ".status")
                notification_settings.set_prop("timeInMinutes", AAZIntType, ".time_in_minutes")
                notification_settings.set_prop("webhookUrl", AAZStrType, ".webhook_url")

            weekly_recurrence = _builder.get(".properties.scheduleParameters[].properties.weeklyRecurrence")
            if weekly_recurrence is not None:
                weekly_recurrence.set_prop("time", AAZStrType, ".time")
                weekly_recurrence.set_prop("weekdays", AAZListType, ".weekdays")

            weekdays = _builder.get(".properties.scheduleParameters[].properties.weeklyRecurrence.weekdays")
            if weekdays is not None:
                weekdays.set_elements(AAZStrType, ".")

            tags = _builder.get(".properties.scheduleParameters[].tags")
            if tags is not None:
                tags.set_elements(AAZStrType, ".")

            tags = _builder.get(".tags")
            if tags is not None:
                tags.set_elements(AAZStrType, ".")

            return self.serialize_content(_content_value)

        def on_200(self, session):
            pass


class _CreateEnvironmentHelper:
    """Helper class for CreateEnvironment"""


__all__ = ["CreateEnvironment"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import time

from azure.cli.testsdk import ScenarioTest, ResourceGroupPreparer, StorageAccountPreparer


class TestLogProfileScenarios(ScenarioTest):

    @ResourceGroupPreparer(name_prefix='cli_test_monitor_lp_create', location='southcentralus')
    @StorageAccountPreparer(location='southcentralus')
    def test_monitor_create_log_profile(self, resource_group, storage_account):

        log_id = self.cmd("monitor log-profiles list --query [0].name -o tsv").output.strip()
        if log_id:
            self.kwargs.update({
                'log_id': log_id
            })
            self.cmd('monitor log-profiles delete -n "{log_id}"')
        self.kwargs.update({
            'name': self.create_random_name('clitest', 20),

        })
        self.kwargs['storage'] = '/subscriptions/{}/resourceGroups/{}/providers/Microsoft.Storage/storageAccounts/{}'\
            .format(self.get_subscription_id(), resource_group, storage_account)
        self.cmd("monitor log-profiles create --categories 'Write' --enabled false --days 100 --location southcentralus --locations westus southcentralus --name {name} --storage-account-id {storage}", checks=[
            self.check('storageAccountId', '{storage}'),
            self.check('serviceBusRuleId', None),
            self.check('retentionPolicy.enabled', 'False')
        ])

        time.sleep(2)
        # test monitor log-profiles show
        self.cmd("monitor log-profiles show -n {name}", checks=[
            self.check('storageAccountId', '{storage}')
        ])

        # test monitor log-profiles update
        self.cmd("monitor log-profiles update -n {name} --set location=southcentralus retentionPolicy.enabled=True",
                 checks=[
                     self.check('retentionPolicy.enabled', 'True')
                 ])

        self.cmd("monitor log-profiles delete -n {name}")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
This script trims Python SDKs by
- Removing aio folders
- Removing unused API versions
"""

import glob
import importlib
import logging
import os
import re
import shutil

import azure.mgmt

from azure.cli.core.profiles import AD_HOC_API_VERSIONS, AZURE_API_PROFILES, ResourceType

_LOGGER = logging.getLogger(__name__)

DRY_RUN = False  # Change to True to dry run


def _rmtree(path):
    _LOGGER.warning(path)
    if not DRY_RUN:
        shutil.rmtree(path)


def calculate_folder_size(start_path):
    """Calculate total size of a folder and file count."""
    # https://stackoverflow.com/questions/1392413/calculating-a-directorys-size-using-python
    total_size = 0
    total_count = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_count += 1
                total_size += os.path.getsize(fp)

    return total_size, total_count


def remove_aio_folders():
    _LOGGER.info("Removing aio folders:")
    mgmt_sdk_dir = azure.mgmt.__path__[0]
    for aio_folder in glob.glob(os.path.join(mgmt_sdk_dir, '**/aio'), recursive=True):
        _rmtree(aio_folder)


def remove_unused_api_versions(resource_type):
    _LOGGER.info(f"Removing unused api folders for {resource_type.import_prefix}:")
    try:
        sdk_path = importlib.import_module(resource_type.import_prefix).__path__[0]
    except ImportError:
        _LOGGER.info(f'{resource_type} is not installed, skip')
        return

    used_api_versions = set()

    # Hard-coded API versions
    if resource_type in AD_HOC_API_VERSIONS:
        ad_hoc_api_versions = set(AD_HOC_API_VERSIONS[resource_type].values())
        used_api_versions.update(ad_hoc_api_versions)

    # API versions in profile
    for profile in AZURE_API_PROFILES.values():
        if resource_type in profile:
            # value is str like '2022-01-01' or SDKProfile
            value = profile[resource_type]
            if value is None:
                _LOGGER.info(f'{resource_type}\'s API version is None, skip')
                return
            if isinstance(value, str):
                used_api_versions.add(value)
            else:
                # SDKProfile
                # default_api_version is in value.profile[None]
                used_api_versions.update(value.profile.values())

    # Convert API version to its folder format: 2019-02-01 -> v2019_02_01
    used_api_folders = {f"v{api.replace('-','_')}" for api in used_api_versions}

    # SDK has a set of versions imported in models.py to form all latest models
    model_file = os.path.join(sdk_path, 'models.py')
    if os.path.exists(model_file):
        with open(model_file, 'r', encoding='utf-8') as f:
            content = f.read()
        for m in re.finditer(r'from \.(v[_\d\w]*)\.models import \*', content):
            used_api_folders.add(m.group(1))

    _LOGGER.info(f'Used API folders: {sorted(used_api_folders)}')

    all_api_folders = {d for d in os.listdir(sdk_path) if os.path.isdir(os.path.join(sdk_path, d)) and d.startswith('v')}
    _LOGGER.info(f'All API folders: {sorted(all_api_folders)}')

    remove_api_folders = sorted(all_api_folders - used_api_folders)
    _LOGGER.info(f'API folders to remove: {remove_api_folders}')

    for api_folder in remove_api_folders:
        full_path = os.path.join(sdk_path, api_folder)
        _rmtree(full_path)


def _print_folder_size(folder):
    size, count = calculate_folder_size(folder)
    size_in_mb = size / 1048576  # 1 MB = 1024 * 1024 B = 1048576 B
    _LOGGER.info(f"{size_in_mb:.2f} MB, {count} files")


def _get_all_sdks_to_trim():
    resource_types = [k for k, v in AZURE_API_PROFILES['latest'].items() if k.import_prefix.startswith('azure.mgmt')]
    return resource_types


def main():
    mgmt_sdk_dir = azure.mgmt.__path__[0]

    # Remove aio folders
    _print_folder_size(mgmt_sdk_dir)
    remove_aio_folders()

    _print_folder_size(mgmt_sdk_dir)

    # Removed unused API versions
    resource_types = _get_all_sdks_to_trim()

    for r in resource_types:
        remove_unused_api_versions(r)

    _print_folder_size(mgmt_sdk_dir)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "network dns zone update",
)
class Update(AAZCommand):
    """Update a DNS zone. Does not modify DNS records within the zone.

    :example: Update a DNS zone properties to change the user-defined value of a previously set tag.
        az network dns zone update -g MyResourceGroup -n www.mysite.com --tags CostCenter=Marketing
    """

    _aaz_info = {
        "version": "2023-07-01-preview",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.network/dnszones/{}", "2023-07-01-preview"],
        ]
    }

    AZ_SUPPORT_GENERIC_UPDATE = True

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.if_match = AAZStrArg(
            options=["--if-match"],
            help="The etag of the DNS zone. Omit this value to always overwrite the current zone. Specify the last-seen etag value to prevent accidentally overwriting any concurrent changes.",
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        _args_schema.zone_name = AAZStrArg(
            options=["-n", "--name", "--zone-name"],
            help="The name of the DNS zone (without a terminating dot).",
            required=True,
            id_part="name",
        )

        # define Arg Group "Parameters"

        _args_schema = cls._args_schema
        _args_schema.tags = AAZDictArg(
            options=["--tags"],
            arg_group="Parameters",
            help="Resource tags.",
            nullable=True,
        )

        tags = cls._args_schema.tags
        tags.Element = AAZStrArg(
            nullable=True,
        )

        # define Arg Group "Properties"
        return cls._args_schema

    _args_sub_resource_update = None

    @classmethod
    def _build_args_sub_resource_update(cls, _schema):
        if cls._args_sub_resource_update is not None:
            _schema.id = cls._args_sub_resource_update.id
            return

        cls._args_sub_resource_update = AAZObjectArg(
            nullable=True,
        )

        sub_resource_update = cls._args_sub_resource_update
        sub_resource_update.id = AAZStrArg(
            options=["id"],
            help="Resource Id.",
            nullable=True,
        )

        _schema.id = cls._args_sub_resource_update.id

    def _execute_operations(self):
        self.pre_operations()
        self.ZonesGet(ctx=self.ctx)()
        self.pre_instance_update(self.ctx.vars.instance)
        self.InstanceUpdateByJson(ctx=self.ctx)()
        self.InstanceUpdateByGeneric(ctx=self.ctx)()
        self.post_instance_update(self.ctx.vars.instance)
        self.ZonesCreateOrUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    @register_callback
    def pre_instance_update(self, instance):
        pass

    @register_callback
    def post_instance_update(self, instance):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class ZonesGet(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Network/dnsZones/{zoneName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
                **self.serialize_url_param(
                    "zoneName", self.ctx.args.zone_name,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2023-07-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()
            _UpdateHelper._build_schema_zone_read(cls._schema_on_200)

            return cls._schema_on_200

    class ZonesCreateOrUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200, 201]:
                return self.on_200_201(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Network/dnsZones/{zoneName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
                **self.serialize_url_param(
                    "zoneName", self.ctx.args.zone_name,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2023-07-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "If-Match", self.ctx.args.if_match,
                ),
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                value=self.ctx.vars.instance,
            )

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()
            _UpdateHelper._build_schema_zone_read(cls._schema_on_200_201)

            return cls._schema_on_200_201

    class InstanceUpdateByJson(AAZJsonInstanceUpdateOperation):

        def __call__(self, *args, **kwargs):
            self._update_instance(self.ctx.vars.instance)

        def _update_instance(self, instance):
            _instance_value, _builder = self.new_content_builder(
                self.ctx.args,
                value=instance,
                typ=AAZObjectType
            )
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})
            _builder.set_prop("tags", AAZDictType, ".tags")

            tags = _builder.get(".tags")
            if tags is not None:
                tags.set_elements(AAZStrType, ".")

            return _instance_value

    class InstanceUpdateByGeneric(AAZGenericInstanceUpdateOperation):

        def __call__(self, *args, **kwargs):
            self._update_instance_by_generic(
                self.ctx.vars.instance,
                self.ctx.generic_update_args
            )


class _UpdateHelper:
    """Helper class for Update"""

    @classmethod
    def _build_schema_sub_resource_update(cls, _builder):
        if _builder is None:
            return
        _builder.set_prop("id", AAZStrType, ".id")

    _schema_sub_resource_read = None

    @classmethod
    def _build_schema_sub_resource_read(cls, _schema):
        if cls._schema_sub_resource_read is not None:
            _schema.id = cls._schema_sub_resource_read.id
            return

        cls._schema_sub_resource_read = _schema_sub_resource_read = AAZObjectType()

        sub_resource_read = _schema_sub_resource_read
        sub_resource_read.id = AAZStrType()

        _schema.id = cls._schema_sub_resource_read.id

    _schema_zone_read = None

    @classmethod
    def _build_schema_zone_read(cls, _schema):
        if cls._schema_zone_read is not None:
            _schema.etag = cls._schema_zone_read.etag
            _schema.id = cls._schema_zone_read.id
            _schema.location = cls._schema_zone_read.location
            _schema.name = cls._schema_zone_read.name
            _schema.properties = cls._schema_zone_read.properties
            _schema.system_data = cls._schema_zone_read.system_data
            _schema.tags = cls._schema_zone_read.tags
            _schema.type = cls._schema_zone_read.type
            return

        cls._schema_zone_read = _schema_zone_read = AAZObjectType()

        zone_read = _schema_zone_read
        zone_read.etag = AAZStrType()
        zone_read.id = AAZStrType(
            flags={"read_only": True},
        )
        zone_read.location = AAZStrType(
            flags={"required": True},
        )
        zone_read.name = AAZStrType(
            flags={"read_only": True},
        )
        zone_read.properties = AAZObjectType(
            flags={"client_flatten": True},
        )
        zone_read.system_data = AAZObjectType(
            serialized_name="systemData",
            flags={"read_only": True},
        )
        zone_read.tags = AAZDictType()
        zone_read.type = AAZStrType(
            flags={"read_only": True},
        )

        properties = _schema_zone_read.properties
        properties.max_number_of_record_sets = AAZIntType(
            serialized_name="maxNumberOfRecordSets",
            flags={"read_only": True},
        )
        properties.max_number_of_records_per_record_set = AAZIntType(
            serialized_name="maxNumberOfRecordsPerRecordSet",
            flags={"read_only": True},
        )
        properties.name_servers = AAZListType(
            serialized_name="nameServers",
            flags={"read_only": True},
        )
        properties.number_of_record_sets = AAZIntType(
            serialized_name="numberOfRecordSets",
            flags={"read_only": True},
        )
        properties.registration_virtual_networks = AAZListType(
            serialized_name="registrationVirtualNetworks",
        )
        properties.resolution_virtual_networks = AAZListType(
            serialized_name="resolutionVirtualNetworks",
        )
        properties.signing_keys = AAZListType(
            serialized_name="signingKeys",
            flags={"read_only": True},
        )
        properties.zone_type = AAZStrType(
            serialized_name="zoneType",
        )

        name_servers = _schema_zone_read.properties.name_servers
        name_servers.Element = AAZStrType()

        registration_virtual_networks = _schema_zone_read.properties.registration_virtual_networks
        registration_virtual_networks.Element = AAZObjectType()
        cls._build_schema_sub_resource_read(registration_virtual_networks.Element)

        resolution_virtual_networks = _schema_zone_read.properties.resolution_virtual_networks
        resolution_virtual_networks.Element = AAZObjectType()
        cls._build_schema_sub_resource_read(resolution_virtual_networks.Element)

        signing_keys = _schema_zone_read.properties.signing_keys
        signing_keys.Element = AAZObjectType()

        _element = _schema_zone_read.properties.signing_keys.Element
        _element.delegation_signer_info = AAZListType(
            serialized_name="delegationSignerInfo",
            flags={"read_only": True},
        )
        _element.flags = AAZIntType(
            flags={"read_only": True},
        )
        _element.key_tag = AAZIntType(
            serialized_name="keyTag",
            flags={"read_only": True},
        )
        _element.protocol = AAZIntType(
            flags={"read_only": True},
        )
        _element.public_key = AAZStrType(
            serialized_name="publicKey",
            flags={"read_only": True},
        )
        _element.security_algorithm_type = AAZIntType(
            serialized_name="securityAlgorithmType",
        )

        delegation_signer_info = _schema_zone_read.properties.signing_keys.Element.delegation_signer_info
        delegation_signer_info.Element = AAZObjectType()

        _element = _schema_zone_read.properties.signing_keys.Element.delegation_signer_info.Element
        _element.digest_algorithm_type = AAZIntType(
            serialized_name="digestAlgorithmType",
        )
        _element.digest_value = AAZStrType(
            serialized_name="digestValue",
            flags={"read_only": True},
        )
        _element.record = AAZStrType(
            flags={"read_only": True},
        )

        system_data = _schema_zone_read.system_data
        system_data.created_at = AAZStrType(
            serialized_name="createdAt",
        )
        system_data.created_by = AAZStrType(
            serialized_name="createdBy",
        )
        system_data.created_by_type = AAZStrType(
            serialized_name="createdByType",
        )
        system_data.last_modified_at = AAZStrType(
            serialized_name="lastModifiedAt",
        )
        system_data.last_modified_by = AAZStrType(
            serialized_name="lastModifiedBy",
        )
        system_data.last_modified_by_type = AAZStrType(
            serialized_name="lastModifiedByType",
        )

        tags = _schema_zone_read.tags
        tags.Element = AAZStrType()

        _schema.etag = cls._schema_zone_read.etag
        _schema.id = cls._schema_zone_read.id
        _schema.location = cls._schema_zone_read.location
        _schema.name = cls._schema_zone_read.name
        _schema.properties = cls._schema_zone_read.properties
        _schema.system_data = cls._schema_zone_read.system_data
        _schema.tags = cls._schema_zone_read.tags
        _schema.type = cls._schema_zone_read.type


__all__ = ["Update"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import sys

from knack.log import get_logger
from knack.prompting import prompt_pass, NoTTYException
from knack.util import CLIError

from azure.cli.core._profile import Profile
from azure.cli.core.util import in_cloud_console

logger = get_logger(__name__)

cloud_resource_type_mappings = {
    "oss-rdbms": "ossrdbms_resource_id",
    "arm": "active_directory_resource_id",
    "aad-graph": "active_directory_graph_resource_id",
    "ms-graph": "microsoft_graph_resource_id",
    "batch": "batch_resource_id",
    "media": "media_resource_id",
    "data-lake": "active_directory_data_lake_resource_id"
}

_CLOUD_CONSOLE_LOGOUT_WARNING = ("Logout successful. Re-login to your initial Cloud Shell identity with"
                                 " 'az login --identity'. Login with a new identity with 'az login'.")
_CLOUD_CONSOLE_LOGIN_WARNING = ("Cloud Shell is automatically authenticated under the initial account signed-in with."
                                " Run 'az login' only if you need to use a different account")


LOGIN_ANNOUNCEMENT = (
    "[Announcements]\n"
    "With the new Azure CLI login experience, you can select the subscription you want to use more easily. "
    "Learn more about it and its configuration at https://go.microsoft.com/fwlink/?linkid=2271236\n\n"
    "If you encounter any problem, please open an issue at https://aka.ms/azclibug\n")

LOGIN_OUTPUT_WARNING = (
    "[Warning] The login output has been updated. Please be aware that it no longer displays the full list of "
    "available subscriptions by default.\n")

USERNAME_PASSWORD_DEPRECATION_WARNING_AZURE_CLOUD = (
    "Starting September 1, 2025, MFA will be gradually enforced for Azure public cloud. "
    "The authentication with username and password in the command line is not supported with MFA. "
    "Consider using one of the compatible authentication methods. "
    "For more details, see https://go.microsoft.com/fwlink/?linkid=2276314")

USERNAME_PASSWORD_DEPRECATION_WARNING_OTHER_CLOUD = (
    "Using authentication with username and password in the command line is strongly discouraged. "
    "Consider using one of the recommended authentication methods. "
    "For more details, see https://go.microsoft.com/fwlink/?linkid=2276314")


def list_subscriptions(cmd, all=False, refresh=False):  # pylint: disable=redefined-builtin
    """List the imported subscriptions."""
    from azure.cli.core.api import load_subscriptions

    subscriptions = load_subscriptions(cmd.cli_ctx, all_clouds=all, refresh=refresh)
    if not subscriptions:
        logger.warning('Please run "az login" to access your accounts.')
    for sub in subscriptions:
        sub['cloudName'] = sub.pop('environmentName', None)
    if not all:
        enabled_ones = [s for s in subscriptions if s.get('state') == 'Enabled']
        if len(enabled_ones) != len(subscriptions):
            logger.warning("A few accounts are skipped as they don't have 'Enabled' state. "
                           "Use '--all' to display them.")
            subscriptions = enabled_ones
    return subscriptions


def show_subscription(cmd, subscription=None):
    profile = Profile(cli_ctx=cmd.cli_ctx)
    return profile.get_subscription(subscription)


def get_access_token(cmd, subscription=None, resource=None, scopes=None, resource_type=None, tenant=None):
    """
    get AAD token to access to a specified resource.
    Use 'az cloud show' command for other Azure resources
    """
    if resource is None and resource_type:
        endpoints_attr_name = cloud_resource_type_mappings[resource_type]
        resource = getattr(cmd.cli_ctx.cloud.endpoints, endpoints_attr_name)

    profile = Profile(cli_ctx=cmd.cli_ctx)
    creds, subscription, tenant = profile.get_raw_token(subscription=subscription, resource=resource, scopes=scopes,
                                                        tenant=tenant)

    result = {
        'tokenType': creds[0],
        'accessToken': creds[1],
        'expires_on': creds[2]['expires_on'],
        'expiresOn': creds[2]['expiresOn'],
        'tenant': tenant
    }
    if subscription:
        result['subscription'] = subscription

    return result


def set_active_subscription(cmd, subscription):
    """Set the current subscription"""
    profile = Profile(cli_ctx=cmd.cli_ctx)
    if not id:
        raise CLIError('Please provide subscription id or unique name.')
    profile.set_active_subscription(subscription)


def account_clear(cmd):
    """Clear all stored subscriptions. To clear individual, use 'logout'"""
    _remove_adal_token_cache()

    if in_cloud_console():
        logger.warning(_CLOUD_CONSOLE_LOGOUT_WARNING)
    profile = Profile(cli_ctx=cmd.cli_ctx)
    profile.logout_all()


def _select_and_set_active(profile, subscriptions):
    """Launch interactive subscription selector and set the chosen subscription as active."""
    from ._subscription_selector import SubscriptionSelector
    from azure.cli.core._profile import _SUBSCRIPTION_ID

    selected = SubscriptionSelector(subscriptions)()
    profile.set_active_subscription(selected[_SUBSCRIPTION_ID])

    print(LOGIN_ANNOUNCEMENT)
    logger.warning(LOGIN_OUTPUT_WARNING)


# pylint: disable=too-many-branches, too-many-locals
def login(cmd, username=None, password=None, tenant=None, scopes=None, allow_no_subscriptions=False,
          claims_challenge=None,
          # Device code flow
          use_device_code=False,
          # Service principal
          service_principal=None, certificate=None, use_cert_sn_issuer=None, client_assertion=None,
          # Managed identity
          identity=False, client_id=None, object_id=None, resource_id=None,
          # Subscription discovery and default subscription selection control
          skip_subscription_discovery=False, subscription=None):
    """Log in to access Azure subscriptions"""

    # quick argument usage check
    if any([password, service_principal, tenant]) and identity:
        raise CLIError("usage error: '--identity' is not applicable with other arguments")
    if identity and username:
        raise CLIError('Passing the managed identity ID with --username is no longer supported. '
                       'Use --client-id, --object-id or --resource-id instead.')
    if any([password, service_principal, username, identity]) and use_device_code:
        raise CLIError("usage error: '--use-device-code' is not applicable with other arguments")
    if use_cert_sn_issuer and not service_principal:
        raise CLIError("usage error: '--use-sn-issuer' is only applicable with a service principal")
    if service_principal and not username:
        raise CLIError('usage error: --service-principal --username NAME --password SECRET --tenant TENANT')
    if skip_subscription_discovery and not tenant:
        raise CLIError("usage error: '--skip-subscription-discovery' requires '--tenant'")
    if skip_subscription_discovery and subscription:
        from azure.cli.core.util import is_guid
        if not is_guid(subscription):
            raise CLIError("usage error: '--subscription' must be a subscription ID (GUID), not a name, "
                           "when combined with '--skip-subscription-discovery'.")
    if username and not service_principal and not identity:
        if cmd.cli_ctx.cloud.endpoints.active_directory.startswith('https://login.microsoftonline.com'):
            logger.warning(USERNAME_PASSWORD_DEPRECATION_WARNING_AZURE_CLOUD)
        else:
            logger.warning(USERNAME_PASSWORD_DEPRECATION_WARNING_OTHER_CLOUD)

    if claims_challenge:
        from azure.cli.core.util import b64decode
        claims_challenge = b64decode(claims_challenge)

    interactive = False

    profile = Profile(cli_ctx=cmd.cli_ctx)

    if identity:
        if in_cloud_console():
            return profile.login_in_cloud_shell()
        return profile.login_with_managed_identity(
            client_id=client_id, object_id=object_id, resource_id=resource_id,
            allow_no_subscriptions=allow_no_subscriptions)
    if in_cloud_console():  # tell users they might not need login
        logger.warning(_CLOUD_CONSOLE_LOGIN_WARNING)

    if username:
        if not (password or client_assertion or certificate):
            try:
                password = prompt_pass('Password: ')
            except NoTTYException:
                raise CLIError('Please specify both username and password in non-interactive mode.')
    else:
        interactive = True

    if service_principal:
        from azure.cli.core.auth.identity import ServicePrincipalAuth
        password = ServicePrincipalAuth.build_credential(
            client_secret=password,
            certificate=certificate, use_cert_sn_issuer=use_cert_sn_issuer,
            client_assertion=client_assertion)

    login_experience_v2 = cmd.cli_ctx.config.getboolean('core', 'login_experience_v2', fallback=True)
    # Send login_experience_v2 config to telemetry
    from azure.cli.core.telemetry import set_login_experience_v2
    set_login_experience_v2(login_experience_v2)

    can_show_selector = (interactive and sys.stdin.isatty() and sys.stdout.isatty() and
                         login_experience_v2)
    is_subscription_filter = bool(subscription or skip_subscription_discovery)

    subscriptions = profile.login(
        interactive,
        username,
        password,
        service_principal,
        tenant,
        scopes=scopes,
        use_device_code=use_device_code,
        allow_no_subscriptions=allow_no_subscriptions,
        use_cert_sn_issuer=use_cert_sn_issuer,
        show_progress=can_show_selector and not skip_subscription_discovery,
        claims_challenge=claims_challenge,
        skip_subscription_discovery=skip_subscription_discovery,
        subscription=subscription
    )

    # Filtered path (--subscription or --skip): show selector only when multiple matches
    if can_show_selector and is_subscription_filter and len(subscriptions) > 1:
        _select_and_set_active(profile, subscriptions)
        return

    # Original no-filter path: always show selector (even 1 item) for backward compatibility
    if can_show_selector and not is_subscription_filter:
        _select_and_set_active(profile, subscriptions)
        return

    all_subscriptions = list(subscriptions)
    for sub in all_subscriptions:
        sub['cloudName'] = sub.pop('environmentName', None)
    return all_subscriptions


def logout(cmd, username=None):
    """Log out to remove access to Azure subscriptions"""
    _remove_adal_token_cache()

    if in_cloud_console():
        logger.warning(_CLOUD_CONSOLE_LOGOUT_WARNING)

    profile = Profile(cli_ctx=cmd.cli_ctx)
    if not username:
        username = profile.get_current_account_user()
    profile.logout(username)


def check_cli(cmd):
    from azure.cli.core.file_util import (
        create_invoker_and_load_cmds_and_args, get_all_help)

    exceptions = {}

    print('Running CLI self-test.\n')

    print('Loading all commands and arguments...')
    try:
        create_invoker_and_load_cmds_and_args(cmd.cli_ctx)
        print('Commands loaded OK.\n')
    except Exception as ex:  # pylint: disable=broad-except
        exceptions['load_commands'] = ex
        logger.error('Error occurred loading commands!\n')
        raise ex

    print('Retrieving all help...')
    try:
        get_all_help(cmd.cli_ctx, skip=False)
        print('Help loaded OK.\n')
    except Exception as ex:  # pylint: disable=broad-except
        exceptions['load_help'] = ex
        logger.error('Error occurred loading help!\n')
        raise ex

    if not exceptions:
        print('CLI self-test completed: OK')
    else:
        raise CLIError(exceptions)


def _remove_adal_token_cache():
    """Remove ADAL token cache file ~/.azure/accessTokens.json, as it is no longer needed by MSAL-based Azure CLI.
    """
    from azure.cli.core._environment import get_config_dir
    adal_token_cache = os.path.join(get_config_dir(), 'accessTokens.json')
    try:
        os.remove(adal_token_cache)
        return True  # Deleted
    except FileNotFoundError:
        return False  # Not exist



# pylint: disable=too-many-lines
# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.command_modules.acs._client_factory import (
    get_container_service_client,
)
from azure.cli.command_modules.acs._consts import (
    CONST_NAMESPACE_ADOPTION_POLICY_ALWAYS,
    CONST_NAMESPACE_ADOPTION_POLICY_IFIDENTICAL,
    CONST_NAMESPACE_ADOPTION_POLICY_NEVER,
    CONST_NAMESPACE_DELETE_POLICY_DELETE,
    CONST_NAMESPACE_DELETE_POLICY_KEEP,
    CONST_NAMESPACE_NETWORK_POLICY_RULE_ALLOWALL,
    CONST_NAMESPACE_NETWORK_POLICY_RULE_ALLOWSAMENAMESPACE,
    CONST_NAMESPACE_NETWORK_POLICY_RULE_DENYALL,
)
from azure.cli.core.azclierror import (
    InvalidArgumentValueError,
    RequiredArgumentMissingError,
)
from azure.cli.core.util import sdk_no_wait
from azure.cli.core.profiles import ResourceType


def get_cluster_location(cmd, resource_group_name, cluster_name):
    containerservice_client = get_container_service_client(cmd.cli_ctx)
    cluster = containerservice_client.managed_clusters.get(resource_group_name, cluster_name)
    return cluster.location


def aks_managed_namespace_add(cmd, client, raw_parameters, headers, no_wait):
    resource_group_name = raw_parameters.get("resource_group_name")
    cluster_name = raw_parameters.get("cluster_name")
    namespace_name = raw_parameters.get("name")

    namespace_config = constructNamespace(cmd, raw_parameters, namespace_name)
    namespace_config.location = get_cluster_location(cmd, resource_group_name, cluster_name)

    return sdk_no_wait(
        no_wait,
        client.begin_create_or_update,
        resource_group_name,
        cluster_name,
        namespace_name,
        namespace_config,
        headers=headers,
    )


def constructNamespace(cmd, raw_parameters, namespace_name):
    tags = raw_parameters.get("tags", {})
    labels_raw = raw_parameters.get("labels")
    labels = parse_key_value_list(labels_raw)
    annotations_raw = raw_parameters.get("annotations")
    annotations = parse_key_value_list(annotations_raw)

    NamespaceProperties = cmd.get_models(
        "NamespaceProperties",
        resource_type=ResourceType.MGMT_CONTAINERSERVICE,
        operation_group="managed_namespaces"
    )

    namespace_properties = NamespaceProperties(
        labels=labels,
        annotations=annotations,
        default_resource_quota=setResourceQuota(cmd, raw_parameters),
        default_network_policy=setNetworkPolicyRule(cmd, raw_parameters),
        adoption_policy=setAdoptionPolicy(raw_parameters),
        delete_policy=setDeletePolicy(raw_parameters)
    )

    Namespace = cmd.get_models(
        "ManagedNamespace",
        resource_type=ResourceType.MGMT_CONTAINERSERVICE,
        operation_group="managed_namespaces"
    )

    namespace_config = Namespace()
    namespace_config.name = namespace_name
    namespace_config.tags = tags
    namespace_config.properties = namespace_properties
    return namespace_config


def setResourceQuota(cmd, raw_parameters):
    cpu_request = raw_parameters.get("cpu_request")
    cpu_limit = raw_parameters.get("cpu_limit")
    memory_request = raw_parameters.get("memory_request")
    memory_limit = raw_parameters.get("memory_limit")

    if any(param is None for param in [cpu_request, cpu_limit, memory_request, memory_limit]):
        raise RequiredArgumentMissingError(
            "Please specify --cpu-request, --cpu-limit, --memory-request, and --memory-limit."
        )

    ResourceQuota = cmd.get_models(
        "ResourceQuota",
        resource_type=ResourceType.MGMT_CONTAINERSERVICE,
        operation_group="managed_namespaces"
    )

    rq = ResourceQuota(
        cpu_request=cpu_request,
        cpu_limit=cpu_limit,
        memory_request=memory_request,
        memory_limit=memory_limit
    )

    return rq


def setNetworkPolicyRule(cmd, raw_parameters):
    ingress_policy = raw_parameters.get("ingress_policy") or CONST_NAMESPACE_NETWORK_POLICY_RULE_ALLOWSAMENAMESPACE
    egress_policy = raw_parameters.get("egress_policy") or CONST_NAMESPACE_NETWORK_POLICY_RULE_ALLOWALL

    valid_network_policy_rules = {
        CONST_NAMESPACE_NETWORK_POLICY_RULE_DENYALL,
        CONST_NAMESPACE_NETWORK_POLICY_RULE_ALLOWSAMENAMESPACE,
        CONST_NAMESPACE_NETWORK_POLICY_RULE_ALLOWALL
    }

    if ingress_policy not in valid_network_policy_rules:
        raise InvalidArgumentValueError(
            f"Invalid ingress_policy '{ingress_policy}'. Must be one of: "
            f"{', '.join(valid_network_policy_rules)}"
        )

    if egress_policy not in valid_network_policy_rules:
        raise InvalidArgumentValueError(
            f"Invalid egress_policy '{egress_policy}'. Must be one of: "
            f"{', '.join(valid_network_policy_rules)}"
        )

    NetworkPolicies = cmd.get_models(
        "NetworkPolicies",
        resource_type=ResourceType.MGMT_CONTAINERSERVICE,
        operation_group="managed_namespaces"
    )

    np = NetworkPolicies(
        ingress=ingress_policy,
        egress=egress_policy
    )

    return np


def setAdoptionPolicy(raw_parameters):
    adoption_policy = raw_parameters.get("adoption_policy") or CONST_NAMESPACE_ADOPTION_POLICY_NEVER

    valid_adoption_policy = {
        CONST_NAMESPACE_ADOPTION_POLICY_NEVER,
        CONST_NAMESPACE_ADOPTION_POLICY_IFIDENTICAL,
        CONST_NAMESPACE_ADOPTION_POLICY_ALWAYS
    }

    if adoption_policy not in valid_adoption_policy:
        raise InvalidArgumentValueError(
            f"Invalid adoption policy '{adoption_policy}'. Must be one of: "
            f"{', '.join(valid_adoption_policy)}"
        )

    return adoption_policy


def setDeletePolicy(raw_parameters):
    delete_policy = raw_parameters.get("delete_policy") or CONST_NAMESPACE_DELETE_POLICY_KEEP

    valid_delete_policy = {
        CONST_NAMESPACE_DELETE_POLICY_KEEP,
        CONST_NAMESPACE_DELETE_POLICY_DELETE
    }

    if delete_policy not in valid_delete_policy:
        raise InvalidArgumentValueError(
            f"Invalid delete policy '{delete_policy}'. Must be one of: "
            f"{', '.join(valid_delete_policy)}"
        )

    return delete_policy


def parse_key_value_list(pairs):
    result = {}
    if pairs is None:
        return result
    for pair in pairs:
        if "=" not in pair:
            raise ValueError(f"Invalid format '{pair}'. Expected format key=value.")
        key, value = pair.split("=", 1)
        result[key.strip()] = value.strip()
    return result


def aks_managed_namespace_update(cmd, client, raw_parameters, headers, existedNamespace, no_wait):
    resource_group_name = raw_parameters.get("resource_group_name")
    cluster_name = raw_parameters.get("cluster_name")
    namespace_name = raw_parameters.get("name")

    namespace_config = updateNamespace(cmd, raw_parameters, existedNamespace)
    namespace_config.location = existedNamespace.location

    return sdk_no_wait(
        no_wait,
        client.begin_create_or_update,
        resource_group_name,
        cluster_name,
        namespace_name,
        namespace_config,
        headers=headers,
    )


def updateNamespace(cmd, raw_parameters, existedNamespace):
    tags = raw_parameters.get("tags", {})
    labels_raw = raw_parameters.get("labels")
    labels = parse_key_value_list(labels_raw)
    annotations_raw = raw_parameters.get("annotations")
    annotations = parse_key_value_list(annotations_raw)

    NamespaceProperties = cmd.get_models(
        "NamespaceProperties",
        resource_type=ResourceType.MGMT_CONTAINERSERVICE,
        operation_group="managed_namespaces"
    )

    namespace_properties = NamespaceProperties(
        labels=labels,
        annotations=annotations,
        default_resource_quota=updateResourceQuota(cmd, raw_parameters, existedNamespace),
        default_network_policy=updateNetworkPolicyRule(cmd, raw_parameters, existedNamespace),
        adoption_policy=updateAdoptionPolicy(raw_parameters, existedNamespace),
        delete_policy=updateDeletePolicy(raw_parameters, existedNamespace)
    )

    Namespace = cmd.get_models(
        "ManagedNamespace",
        resource_type=ResourceType.MGMT_CONTAINERSERVICE,
        operation_group="managed_namespaces"
    )

    namespace_config = Namespace()
    namespace_config.name = existedNamespace.name
    namespace_config.tags = tags
    namespace_config.properties = namespace_properties
    return namespace_config


def updateResourceQuota(cmd, raw_parameters, existedNamespace):
    cpu_request = raw_parameters.get("cpu_request")
    cpu_limit = raw_parameters.get("cpu_limit")
    memory_request = raw_parameters.get("memory_request")
    memory_limit = raw_parameters.get("memory_limit")

    if cpu_request is None:
        cpu_request = existedNamespace.properties.default_resource_quota.cpu_request

    if cpu_limit is None:
        cpu_limit = existedNamespace.properties.default_resource_quota.cpu_limit

    if memory_request is None:
        memory_request = existedNamespace.properties.default_resource_quota.memory_request

    if memory_limit is None:
        memory_limit = existedNamespace.properties.default_resource_quota.memory_limit

    ResourceQuota = cmd.get_models(
        "ResourceQuota",
        resource_type=ResourceType.MGMT_CONTAINERSERVICE,
        operation_group="managed_namespaces"
    )

    rq = ResourceQuota(
        cpu_request=cpu_request,
        cpu_limit=cpu_limit,
        memory_request=memory_request,
        memory_limit=memory_limit
    )

    return rq


def updateNetworkPolicyRule(cmd, raw_parameters, existedNamespace):
    ingress_policy = raw_parameters.get("ingress_policy")
    egress_policy = raw_parameters.get("egress_policy")

    valid_network_policy_rules = {
        CONST_NAMESPACE_NETWORK_POLICY_RULE_DENYALL,
        CONST_NAMESPACE_NETWORK_POLICY_RULE_ALLOWSAMENAMESPACE,
        CONST_NAMESPACE_NETWORK_POLICY_RULE_ALLOWALL
    }

    if ingress_policy is not None and ingress_policy not in valid_network_policy_rules:
        raise InvalidArgumentValueError(
            f"Invalid ingress_policy '{ingress_policy}'. Must be one of: "
            f"{', '.join(valid_network_policy_rules)}"
        )

    if egress_policy is not None and egress_policy not in valid_network_policy_rules:
        raise InvalidArgumentValueError(
            f"Invalid egress_policy '{egress_policy}'. Must be one of: "
            f"{', '.join(valid_network_policy_rules)}"
        )

    if ingress_policy is None:
        ingress_policy = existedNamespace.properties.default_network_policy.ingress

    if egress_policy is None:
        egress_policy = existedNamespace.properties.default_network_policy.egress

    NetworkPolicies = cmd.get_models(
        "NetworkPolicies",
        resource_type=ResourceType.MGMT_CONTAINERSERVICE,
        operation_group="managed_namespaces"
    )

    np = NetworkPolicies(
        ingress=ingress_policy,
        egress=egress_policy
    )

    return np


def updateAdoptionPolicy(raw_parameters, existedNamespace):
    adoption_policy = raw_parameters.get("adoption_policy")

    valid_adoption_policy = {
        CONST_NAMESPACE_ADOPTION_POLICY_NEVER,
        CONST_NAMESPACE_ADOPTION_POLICY_IFIDENTICAL,
        CONST_NAMESPACE_ADOPTION_POLICY_ALWAYS
    }

    if adoption_policy is not None and adoption_policy not in valid_adoption_policy:
        raise InvalidArgumentValueError(
            f"Invalid adoption policy '{adoption_policy}'. Must be one of: "
            f"{', '.join(valid_adoption_policy)}"
        )

    if adoption_policy is None:
        adoption_policy = existedNamespace.properties.adoption_policy

    return adoption_policy


def updateDeletePolicy(raw_parameters, existedNamespace):
    delete_policy = raw_parameters.get("delete_policy")

    valid_delete_policy = {
        CONST_NAMESPACE_DELETE_POLICY_KEEP,
        CONST_NAMESPACE_DELETE_POLICY_DELETE
    }

    if delete_policy is not None and delete_policy not in valid_delete_policy:
        raise InvalidArgumentValueError(
            f"Invalid delete policy '{delete_policy}'. Must be one of: "
            f"{', '.join(valid_delete_policy)}"
        )

    if delete_policy is None:
        delete_policy = existedNamespace.properties.delete_policy

    return delete_policy



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.testsdk import ScenarioTest, ResourceGroupPreparer, StorageAccountPreparer
from azure.cli.testsdk.scenario_tests import AllowLargeResponse


# This is a live test because the start and end time can't be determined dynamically
class TestActivityLogScenarios(ScenarioTest):

    @AllowLargeResponse()
    @ResourceGroupPreparer(location='southcentralus')
    @StorageAccountPreparer()
    def test_activity_log_list_scenario(self, resource_group):
        from azure.core.exceptions import HttpResponseError
        with self.assertRaisesRegex(HttpResponseError, 'start time cannot be more than 90 days in the past'):
            self.cmd('monitor activity-log list --start-time 2018-01-01T00:00:00Z --end-time 2999-01-01T00:00:00Z')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger
from knack.util import CLIError

logger = get_logger(__name__)

DEFAULT_INDEX_URL = "https://aka.ms/azure-cli-extension-index-v1"

ERR_TMPL_EXT_INDEX = 'Unable to get extension index.\n'
ERR_TMPL_NON_200 = '{}Server returned status code {{}} for {{}}'.format(ERR_TMPL_EXT_INDEX)
ERR_TMPL_NO_NETWORK = '{}Please ensure you have network connection. Error detail: {{}}'.format(ERR_TMPL_EXT_INDEX)
ERR_TMPL_BAD_JSON = '{}Response body does not contain valid json. Error detail: {{}}'.format(ERR_TMPL_EXT_INDEX)

ERR_UNABLE_TO_GET_EXTENSIONS = 'Unable to get extensions from index. Improper index format.'
TRIES = 3


def get_index_url(cli_ctx=None):
    """Use extension index url in the order of:
    1. Environment variable: AZURE_EXTENSION_INDEX_URL
    2. Config setting: extension.index_url
    3. Index file in azmirror storage account cloud endpoint
    4. DEFAULT_INDEX_URL
    """
    import posixpath
    if cli_ctx:
        url = cli_ctx.config.get('extension', 'index_url', None)
        if url:
            return url
    azmirror_endpoint = cli_ctx.cloud.endpoints.azmirror_storage_account_resource_id if cli_ctx and \
        cli_ctx.cloud.endpoints.has_endpoint_set('azmirror_storage_account_resource_id') else None
    return posixpath.join(azmirror_endpoint, 'extensions', 'index.json') if azmirror_endpoint else DEFAULT_INDEX_URL


# pylint: disable=inconsistent-return-statements
def get_index(index_url=None, cli_ctx=None):
    import requests
    from azure.cli.core.util import should_disable_connection_verify
    index_url = index_url or get_index_url(cli_ctx=cli_ctx)

    for try_number in range(TRIES):
        try:
            response = requests.get(index_url, verify=not should_disable_connection_verify())
            if response.status_code == 200:
                return response.json()
            msg = ERR_TMPL_NON_200.format(response.status_code, index_url)
            raise CLIError(msg)
        except (requests.exceptions.ConnectionError, requests.exceptions.HTTPError, ValueError) as err:
            if try_number == TRIES - 1:
                # ValueError indicates that shortlink url is not redirecting properly to intended index url
                msg = ERR_TMPL_BAD_JSON.format(str(err)) if isinstance(err, ValueError) else \
                    ERR_TMPL_NO_NETWORK.format(str(err))
                raise CLIError(msg)
            import time
            time.sleep(0.5)
            continue


def get_index_extensions(index_url=None, cli_ctx=None):
    index = get_index(index_url=index_url, cli_ctx=cli_ctx)
    extensions = index.get('extensions')
    if extensions is None:
        logger.warning(ERR_UNABLE_TO_GET_EXTENSIONS)
    return extensions



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import re
from knack.util import CLIError
from knack.log import get_logger
from azure.cli.core.azclierror import FileOperationError, InvalidArgumentValueError
from ._constants import ACR_NAME_VALIDATION_REGEX, USER_ASSIGNED_IDENTITY_RESOURCE_ID_TEMPLATE

BAD_REPO_FQDN = "The positional parameter 'repo_id' must be a fully qualified repository specifier such"\
                " as 'myregistry.azurecr.io/hello-world'."
BAD_PERM_REPO_FQDN = "The positional parameter 'perm_repo_id' must be a fully qualified repository specifier such"\
                     " as 'myregistry.azurecr.io/hello-world'. It may optionally specify a"\
                     " tag such as 'myregistry.azurecr.io/hello-world:latest'."
BAD_MANIFEST_FQDN = "The positional parameter 'manifest_id' must be a fully qualified"\
                    " manifest specifier such as 'myregistry.azurecr.io/hello-world:latest' or"\
                    " 'myregistry.azurecr.io/hello-world@sha256:abc123'."
BAD_REGISTRY_NAME = "Registry names may contain only alpha numeric characters and must be between 5 and 50 characters"
INVALID_LOGIN_SERVER_SUFFIX = "The login server suffix is not valid for the current cloud. Please try again using"\
                              " '{}'."

logger = get_logger(__name__)


def validate_headers(namespace):
    """Extracts multiple space-separated headers in key[=value] format. """
    if isinstance(namespace.headers, list):
        headers_dict = {}
        for item in namespace.headers:
            headers_dict.update(validate_header(item))
        namespace.headers = headers_dict


def validate_header(string):
    """Extracts a single header in key[=value] format. """
    result = {}
    if string:
        comps = string.split('=', 1)
        result = {comps[0]: comps[1]} if len(comps) > 1 else {string: ''}
    return result


def validate_arg(namespace):
    if isinstance(namespace.arg, list):
        arguments_list = []
        for item in namespace.arg:
            arguments_list.append(validate_task_argument(item, False))
        namespace.arg = arguments_list


def validate_secret_arg(namespace):
    if isinstance(namespace.secret_arg, list):
        secret_arguments_list = []
        for item in namespace.secret_arg:
            secret_arguments_list.append(validate_task_argument(item, True))
        namespace.secret_arg = secret_arguments_list


def validate_set(namespace):
    if isinstance(namespace.set_value, list):
        set_list = []
        for item in namespace.set_value:
            set_list.append(validate_task_value(item, False))
        namespace.set_value = set_list


def validate_set_secret(namespace):
    if isinstance(namespace.set_secret, list):
        set_secret_list = []
        for item in namespace.set_secret:
            set_secret_list.append(validate_task_value(item, True))
        namespace.set_secret = set_secret_list


def validate_task_value(string, is_secret):
    """Extracts a single SetValue in key[=value] format. """
    if string:
        comps = string.split('=', 1)
        if len(comps) > 1:
            return {'type': 'SetValue', 'name': comps[0], 'value': comps[1], 'isSecret': is_secret}
        return {'type': 'SetValue', 'name': comps[0], 'value': '', 'isSecret': is_secret}
    return None


def validate_task_argument(string, is_secret):
    """Extracts a single argument in key[=value] format. """
    if string:
        comps = string.split('=', 1)
        if len(comps) > 1:
            return {'type': 'Argument', 'name': comps[0], 'value': comps[1], 'isSecret': is_secret}
        # If no value, check if the argument exists as an environment variable
        local_value = os.environ.get(comps[0])
        if local_value is not None:
            return {'type': 'Argument', 'name': comps[0], 'value': local_value, 'isSecret': is_secret}
        return {'type': 'Argument', 'name': comps[0], 'value': '', 'isSecret': is_secret}
    return None


def validate_retention_days(namespace):
    days = namespace.days
    if days and (days < 0 or days > 365):
        raise CLIError("Invalid value for days: should be from 0 to 365")


def validate_registry_name(cmd, namespace):
    """Omit login server endpoint suffix and domain name label (DNL) hash if given."""
    registry = namespace.registry_name
    if registry is None:
        return
    suffixes = cmd.cli_ctx.cloud.suffixes

    # Split registry login server into components ['myregistry-dnlhash', '.azurecr.io']
    registry_parts = registry.split('.', 1)
    trimmed_registry_name = registry_parts[0]
    registry_login_server_suffix = '.' + registry_parts[1] if len(registry_parts) > 1 else ''

    dnl_hash_index = trimmed_registry_name.find("-")

    # Registry name has hyphen but no login server endpoint suffix
    if registry_login_server_suffix == '' and dnl_hash_index != -1:
        raise InvalidArgumentValueError(BAD_REGISTRY_NAME)

    # Some clouds do not define 'acr_login_server_endpoint' (e.g. AzureGermanCloud)
    if hasattr(suffixes, 'acr_login_server_endpoint'):
        acr_suffix = suffixes.acr_login_server_endpoint
        if registry_login_server_suffix.lower() == acr_suffix and registry_login_server_suffix != '':
            if dnl_hash_index != -1:
                removed_suffix = trimmed_registry_name[dnl_hash_index:] + registry_login_server_suffix
                registry_name = trimmed_registry_name[:dnl_hash_index]
            else:
                removed_suffix = registry_login_server_suffix
                registry_name = trimmed_registry_name
            logger.warning("Registry name is '%s'. The following suffix '%s' is automatically omitted.",
                           registry_name,
                           removed_suffix)
        else:
            if registry_login_server_suffix != '':
                raise InvalidArgumentValueError(INVALID_LOGIN_SERVER_SUFFIX.format(acr_suffix))
            registry_name = trimmed_registry_name
        namespace.registry_name = registry_name
        registry = registry_name

    registry = namespace.registry_name
    if not re.match(ACR_NAME_VALIDATION_REGEX, registry):
        raise InvalidArgumentValueError(BAD_REGISTRY_NAME)


def validate_expiration_time(namespace):
    import datetime
    DATE_TIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
    if namespace.expiration:
        try:
            namespace.expiration = datetime.datetime.strptime(namespace.expiration, DATE_TIME_FORMAT)
        except ValueError:
            raise CLIError("Input '{}' is not valid datetime. Valid example: 2025-12-31T12:59:59Z".format(
                namespace.expiration))


def validate_repo_id(namespace):
    if namespace.repo_id:
        repo_id = namespace.repo_id[0]
        if '.' not in repo_id or '/' not in repo_id:
            raise InvalidArgumentValueError(BAD_REPO_FQDN)


def validate_permissive_repo_id(namespace):
    if namespace.perm_repo_id:
        perm_repo_id = namespace.perm_repo_id[0]
        if '.' not in perm_repo_id or '/' not in perm_repo_id:
            raise InvalidArgumentValueError(BAD_PERM_REPO_FQDN)
        if '@' in perm_repo_id:
            raise InvalidArgumentValueError(BAD_PERM_REPO_FQDN)


def validate_manifest_id(namespace):
    if namespace.manifest_id:
        manifest_id = namespace.manifest_id[0]
        if '.' not in manifest_id or '/' not in manifest_id:
            raise InvalidArgumentValueError(BAD_MANIFEST_FQDN)


def validate_repository(namespace):
    if namespace.repository:
        if ':' in namespace.repository:
            raise InvalidArgumentValueError("Parameter 'name' refers to a repository and"
                                            " should not include a tag or digest.")


def validate_docker_file_path(docker_file_path):
    if not os.path.isfile(docker_file_path):
        raise FileOperationError("Unable to find '{}'.".format(docker_file_path))


def validate_cache_credentials(namespace):
    """Validate cache credential options - allow both --identity and --cred-set, but --remove-cred-set is exclusive."""
    has_identity = namespace.identity is not None
    has_cred_set = namespace.cred_set is not None
    has_remove_cred_set = getattr(namespace, 'remove_cred_set', False)

    if has_remove_cred_set and (has_identity or has_cred_set):
        raise InvalidArgumentValueError(
            "Cannot specify --remove-cred-set with other credential options. "
            "Use --remove-cred-set alone to remove credentials."
        )

    # Validate identity format if provided
    if has_identity:
        identity_pattern = (
            r'^/subscriptions/[^/]+/resource[Gg]roups/[^/]+'
            r'/providers/Microsoft\.ManagedIdentity'
            r'/userAssignedIdentities/[^/]+$'
        )

        if not re.match(identity_pattern, namespace.identity, re.IGNORECASE):
            example_format = USER_ASSIGNED_IDENTITY_RESOURCE_ID_TEMPLATE.format(
                sub_id='{subscriptionId}',
                rg='{resourceGroupName}',
                identity_name='{identityName}'
            )
            raise InvalidArgumentValueError(
                f"The --identity parameter must be a valid ARM resource ID "
                f"for a user-assigned managed identity. "
                f"Format: {example_format}"
            )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands.client_factory import get_mgmt_service_client


def get_acr_service_client(cli_ctx, api_version=None):
    """Returns the client for managing container registries. """
    from azure.cli.core.profiles import ResourceType
    return get_mgmt_service_client(cli_ctx, ResourceType.MGMT_CONTAINERREGISTRY, api_version=api_version)


def get_acr_tasks_service_client(cli_ctx, api_version=None):
    """Returns the client for managing container registry tasks."""
    from azure.cli.core.profiles import ResourceType
    return get_mgmt_service_client(cli_ctx, ResourceType.MGMT_CONTAINERREGISTRYTASKS, api_version=api_version)


# The function is used in Azure and Edge and hybrid profile is used to support the different API versions.
def cf_acr_registries(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).registries


def cf_acr_cache(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).cache_rules


def cf_acr_cred_sets(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).credential_sets


def cf_acr_registries_tasks(cli_ctx, *_):
    return get_acr_tasks_service_client(cli_ctx).registries


def cf_acr_replications(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).replications


def cf_acr_webhooks(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).webhooks


def cf_acr_private_endpoint_connections(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).private_endpoint_connections


def cf_acr_tasks(cli_ctx, *_):
    return get_acr_tasks_service_client(cli_ctx).tasks


def cf_acr_taskruns(cli_ctx, *_):
    return get_acr_tasks_service_client(cli_ctx).task_runs


def cf_acr_runs(cli_ctx, *_):
    return get_acr_tasks_service_client(cli_ctx).runs


def cf_acr_scope_maps(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).scope_maps


def cf_acr_tokens(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).tokens


def cf_acr_token_credentials(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).registries


def cf_acr_agentpool(cli_ctx, *_):
    return get_acr_tasks_service_client(cli_ctx).agent_pools


def cf_acr_connected_registries(cli_ctx, *_):
    return get_acr_service_client(cli_ctx).connected_registries



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import contextlib
import functools

from .base import ReplayableTest
from .utilities import create_random_name, is_text_payload, trim_kwargs_from_test_function
from .recording_processors import RecordingProcessor


# Core Utility

class AbstractPreparer:
    def __init__(self, name_prefix, name_len, disable_recording=False):
        self.name_prefix = name_prefix
        self.name_len = name_len
        self.resource_moniker = None
        self.resource_random_name = None
        self.test_class_instance = None
        self.live_test = False
        self.disable_recording = disable_recording

    def __call__(self, fn):
        def _preparer_wrapper(test_class_instance, **kwargs):
            self.live_test = not isinstance(test_class_instance, ReplayableTest)
            self.test_class_instance = test_class_instance

            if self.live_test or test_class_instance.in_recording:
                resource_name = self.random_name
                if not self.live_test and isinstance(self, RecordingProcessor):
                    test_class_instance.recording_processors.append(self)
            else:
                resource_name = self.moniker

            with self.override_disable_recording():
                parameter_update = self.create_resource(
                    resource_name,
                    **kwargs
                )

            if parameter_update:
                kwargs.update(parameter_update)

            trim_kwargs_from_test_function(fn, kwargs)

            try:
                fn(test_class_instance, **kwargs)
            finally:
                # Russian Doll - the last declared resource to be deleted first.
                self.remove_resource_with_record_override(resource_name, **kwargs)

        setattr(_preparer_wrapper, '__is_preparer', True)
        functools.update_wrapper(_preparer_wrapper, fn)
        return _preparer_wrapper

    @contextlib.contextmanager
    def override_disable_recording(self):
        if hasattr(self.test_class_instance, 'disable_recording'):
            orig_enabled = self.test_class_instance.disable_recording
            self.test_class_instance.disable_recording = self.disable_recording
            yield
            self.test_class_instance.disable_recording = orig_enabled
        else:
            yield

    @property
    def moniker(self):
        if not self.resource_moniker:
            self.test_class_instance.test_resources_count += 1
            self.resource_moniker = '{}{:06}'.format(self.name_prefix,
                                                     self.test_class_instance.test_resources_count)
        return self.resource_moniker

    def create_random_name(self):
        return create_random_name(self.name_prefix, self.name_len)

    @property
    def random_name(self):
        if not self.resource_random_name:
            self.resource_random_name = self.create_random_name()
        return self.resource_random_name

    def create_resource(self, name, **kwargs):  # pylint: disable=unused-argument,no-self-use
        return {}

    def remove_resource(self, name, **kwargs):  # pylint: disable=unused-argument
        pass

    def remove_resource_with_record_override(self, name, **kwargs):
        with self.override_disable_recording():
            self.remove_resource(name, **kwargs)


class SingleValueReplacer(RecordingProcessor):
    # pylint: disable=no-member
    def process_request(self, request):
        from urllib.parse import quote_plus
        if self.random_name in request.uri:
            request.uri = request.uri.replace(self.random_name, self.moniker)
        elif quote_plus(self.random_name) in request.uri:
            request.uri = request.uri.replace(quote_plus(self.random_name),
                                              quote_plus(self.moniker))

        if is_text_payload(request) and request.body:
            body = str(request.body, 'utf-8') if isinstance(request.body, bytes) else str(request.body)
            if self.random_name in body:
                request.body = body.replace(self.random_name, self.moniker)

        return request

    def process_response(self, response):
        if is_text_payload(response) and response['body']['string']:
            response['body']['string'] = response['body']['string'].replace(self.random_name,
                                                                            self.moniker)

        self.replace_header(response, 'location', self.random_name, self.moniker)
        self.replace_header(response, 'azure-asyncoperation', self.random_name, self.moniker)

        return response



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=unused-argument, line-too-long
from functools import cmp_to_key
from importlib import import_module
from urllib.parse import quote
from knack.log import get_logger
from knack.prompting import prompt
from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.core.local_context import ALL
from azure.cli.core.util import CLIError, sdk_no_wait, user_confirmation
from azure.core.exceptions import ResourceNotFoundError
from azure.mgmt import postgresqlflexibleservers as postgresql_flexibleservers
from azure.mgmt.core.tools import is_valid_resource_id, parse_resource_id, resource_id
from .._client_factory import (
    cf_postgres_check_resource_availability,
    cf_postgres_flexible_admin,
    cf_postgres_flexible_config,
    cf_postgres_flexible_db,
    cf_postgres_flexible_firewall_rules,
    cf_postgres_flexible_private_dns_zone_suffix_operations,
    get_postgresql_flexible_management_client)
from .._db_context import DbContext
from ..utils._flexible_server_location_capabilities_util import (
    get_postgres_location_capability_info,
    get_postgres_server_capability_info)
from ..utils._flexible_server_util import (
    _is_resource_name,
    build_identity_and_data_encryption,
    generate_missing_parameters,
    generate_password,
    get_current_time,
    get_postgres_skus,
    get_postgres_tiers,
    parse_maintenance_window,
    resolve_poller)
from ..utils.validators import (
    build_network_configuration,
    check_resource_group,
    compare_sku_names,
    pg_arguments_validator,
    pg_byok_validator,
    pg_restore_validator,
    resolve_private_dns_zone_id,
    validate_and_format_restore_point_in_time,
    validate_citus_cluster,
    validate_georestore_network,
    validate_resource_group,
    validate_server_name)
from .firewall_rule_commands import create_firewall_rule
from .microsoft_entra_commands import _create_admin

logger = get_logger(__name__)
DEFAULT_DB_NAME = 'flexibleserverdb'
POSTGRES_DB_NAME = 'postgres'
DELEGATION_SERVICE_NAME = "Microsoft.DBforPostgreSQL/flexibleServers"
RESOURCE_PROVIDER = 'Microsoft.DBforPostgreSQL'


# region create without args
# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
# pylint: disable=raise-missing-from, unbalanced-tuple-unpacking
def flexible_server_create(cmd, client,
                           resource_group_name=None, server_name=None,
                           location=None, backup_retention=None,
                           sku_name=None, tier=None,
                           storage_gb=None, version=None, microsoft_entra_auth=None,
                           admin_name=None, admin_id=None, admin_type=None,
                           password_auth=None, administrator_login=None, administrator_login_password=None,
                           tags=None, subnet=None, vnet=None,
                           private_dns_zone_arguments=None, public_access=None,
                           zonal_resiliency=None, allow_same_zone=False,
                           zone=None, standby_availability_zone=None,
                           geo_redundant_backup=None, byok_identity=None, byok_key=None, backup_byok_identity=None, backup_byok_key=None,
                           federated_client_id=None, backup_federated_client_id=None,
                           auto_grow=None, performance_tier=None,
                           storage_type=None, iops=None, throughput=None, cluster_size=None, database_name=None, yes=False):

    if not check_resource_group(resource_group_name):
        resource_group_name = None

    # Generate missing parameters
    location, resource_group_name, server_name = generate_missing_parameters(cmd, location, resource_group_name,
                                                                             server_name, 'postgres')

    db_context = DbContext(
        cmd=cmd, azure_sdk=postgresql_flexibleservers, cf_firewall=cf_postgres_flexible_firewall_rules,
        cf_db=cf_postgres_flexible_db, cf_availability=cf_postgres_check_resource_availability,
        cf_private_dns_zone_suffix=cf_postgres_flexible_private_dns_zone_suffix_operations,
        logging_name='PostgreSQL', command_group='postgres', server_client=client, location=location)

    server_name = server_name.lower()
    high_availability_mode = "Disabled"

    if (sku_name is None) or (version is None) or \
       (zonal_resiliency is not None and zonal_resiliency.lower() != 'disabled'):
        list_location_capability_info = get_postgres_location_capability_info(cmd, location)

        # set sku_name from capability API
        if sku_name is None:
            tiers = [item.lower() for item in get_postgres_tiers(list_location_capability_info['sku_info'])]
            try:
                sku_info = list_location_capability_info['sku_info']
                skus = list(get_postgres_skus(sku_info, tier.lower()))
                skus = sorted(skus, key=cmp_to_key(compare_sku_names))
                sku_name = skus[0]
            except:
                raise CLIError('Incorrect value for --tier. Allowed values : {}'.format(tiers))
        # default to the latest version
        if version is None:
            supported_server_versions = sorted(list_location_capability_info['supported_server_versions'])
            version = supported_server_versions[-1]
        # set high availability from capability API
        if (zonal_resiliency is not None and zonal_resiliency.lower() != 'disabled'):
            single_az = list_location_capability_info['single_az']
            high_availability_mode = 'SameZone' if single_az and allow_same_zone else 'ZoneRedundant'

    pg_arguments_validator(db_context,
                           server_name=server_name,
                           location=location,
                           tier=tier, sku_name=sku_name,
                           storage_gb=storage_gb,
                           auto_grow=auto_grow,
                           storage_type=storage_type,
                           iops=iops, throughput=throughput,
                           zonal_resiliency=zonal_resiliency,
                           allow_same_zone=allow_same_zone,
                           standby_availability_zone=standby_availability_zone,
                           zone=zone,
                           subnet=subnet,
                           public_access=public_access,
                           version=version,
                           geo_redundant_backup=geo_redundant_backup,
                           byok_identity=byok_identity,
                           byok_key=byok_key,
                           backup_byok_identity=backup_byok_identity,
                           backup_byok_key=backup_byok_key,
                           federated_client_id=federated_client_id,
                           backup_federated_client_id=backup_federated_client_id,
                           performance_tier=performance_tier,
                           password_auth=password_auth, microsoft_entra_auth=microsoft_entra_auth,
                           admin_name=admin_name, admin_id=admin_id, admin_type=admin_type, database_name=database_name)

    cluster = None
    if cluster_size is not None:
        cluster = postgresql_flexibleservers.models.Cluster(cluster_size=cluster_size, default_database_name=database_name if database_name else POSTGRES_DB_NAME)

    server_result = firewall_id = None

    network, start_ip, end_ip = build_network_configuration(
        cmd=cmd,
        resource_group_name=resource_group_name,
        server_name=server_name,
        location=location,
        db_context=db_context,
        private_dns_zone_arguments=private_dns_zone_arguments,
        public_access=public_access,
        vnet=vnet,
        subnet=subnet,
        yes=yes)

    storage = postgresql_flexibleservers.models.Storage(storage_size_gb=storage_gb, auto_grow=auto_grow, tier=performance_tier, type=storage_type, iops=iops, throughput=throughput)

    backup = postgresql_flexibleservers.models.Backup(backup_retention_days=backup_retention,
                                                      geo_redundant_backup=geo_redundant_backup)

    sku = postgresql_flexibleservers.models.Sku(name=sku_name, tier=tier)

    high_availability = postgresql_flexibleservers.models.HighAvailability(mode=high_availability_mode,
                                                                           standby_availability_zone=standby_availability_zone)

    is_password_auth_enabled = bool(password_auth is not None and password_auth.lower() == 'enabled')
    is_microsoft_entra_auth_enabled = bool(microsoft_entra_auth is not None and microsoft_entra_auth.lower() == 'enabled')
    if is_password_auth_enabled:
        administrator_login_password = generate_password(administrator_login_password)

    identity, data_encryption = build_identity_and_data_encryption(byok_identity=byok_identity,
                                                                   byok_key=byok_key,
                                                                   backup_byok_identity=backup_byok_identity,
                                                                   backup_byok_key=backup_byok_key,
                                                                   federated_client_id=federated_client_id,
                                                                   backup_federated_client_id=backup_federated_client_id)

    auth_config = postgresql_flexibleservers.models.AuthConfig(active_directory_auth='Enabled' if is_microsoft_entra_auth_enabled else 'Disabled',
                                                               password_auth=password_auth)

    # Create postgresql
    # Note : passing public_access has no effect as the accepted values are 'Enabled' and 'Disabled'. So the value ends up being ignored.
    server_result = _create_server(db_context, cmd, resource_group_name, server_name,
                                   tags=tags,
                                   location=location,
                                   sku=sku,
                                   administrator_login=administrator_login,
                                   administrator_login_password=administrator_login_password,
                                   storage=storage,
                                   backup=backup,
                                   network=network,
                                   version=version,
                                   high_availability=high_availability,
                                   availability_zone=zone,
                                   identity=identity,
                                   data_encryption=data_encryption,
                                   auth_config=auth_config,
                                   cluster=cluster)

    # Add Microsoft Entra Admin
    if is_microsoft_entra_auth_enabled and admin_name is not None or admin_id is not None:
        server_admin_client = cf_postgres_flexible_admin(cmd.cli_ctx, '_')
        logger.warning("Add Microsoft Entra Admin '%s'.", admin_name)
        _create_admin(server_admin_client, resource_group_name, server_name, admin_name, admin_id, admin_type)

    # Adding firewall rule
    if start_ip != -1 and end_ip != -1:
        firewall_id = create_firewall_rule(db_context, cmd, resource_group_name, server_name, start_ip, end_ip)

    user = server_result.administrator_login if is_password_auth_enabled else '<user>'
    # [SuppressMessage("Microsoft.Security", "CS002:SecretInNextLine", Justification="Not a hardcoded secret - value is either a runtime-provided variable or the placeholder '<password>'")]
    password = administrator_login_password if is_password_auth_enabled else '<password>'
    server_id = server_result.id
    loc = server_result.location
    version = server_result.version
    sku = server_result.sku.name
    host = server_result.fully_qualified_domain_name
    subnet_id = None if network is None else network.delegated_subnet_resource_id

    if is_password_auth_enabled:
        logger.warning('Make a note of your password. If you forget, you would have to '
                       'reset your password with "az postgres flexible-server update -n %s -g %s -p <new-password>".',
                       server_name, resource_group_name)
    logger.warning('Try using \'az postgres flexible-server connect\' command to test out connection.')

    _update_local_contexts(cmd, server_name, resource_group_name, location, user)

    return _form_response(user, sku, loc, server_id, host, version, password,
                          firewall_id, subnet_id, is_password_auth_enabled, is_microsoft_entra_auth_enabled, admin_name)


def _create_server(db_context, cmd, resource_group_name, server_name, tags, location, sku, administrator_login, administrator_login_password,
                   storage, backup, network, version, high_availability, availability_zone, identity, data_encryption, auth_config, cluster):
    validate_resource_group(resource_group_name)

    logging_name, server_client = db_context.logging_name, db_context.server_client
    logger.warning('Creating %s Server \'%s\' in group \'%s\'...', logging_name, server_name, resource_group_name)

    logger.warning('Your server \'%s\' is using sku \'%s\' (Paid Tier). '
                   'Please refer to https://aka.ms/postgres-pricing for pricing details', server_name, sku.name)

    # Note : passing public-network-access has no effect as the accepted values are 'Enabled' and 'Disabled'.
    # So when you pass an IP here(from the CLI args of public_access), it ends up being ignored.
    parameters = postgresql_flexibleservers.models.Server(
        tags=tags,
        location=location,
        sku=sku,
        administrator_login=administrator_login,
        administrator_login_password=administrator_login_password,
        storage=storage,
        backup=backup,
        network=network,
        version=version,
        high_availability=high_availability,
        availability_zone=availability_zone,
        identity=identity,
        data_encryption=data_encryption,
        auth_config=auth_config,
        cluster=cluster,
        create_mode="Create")

    return resolve_poller(
        server_client.begin_create_or_update(resource_group_name, server_name, parameters), cmd.cli_ctx,
        '{} Server Create'.format(logging_name))


def _update_local_contexts(cmd, server_name, resource_group_name, location, user):
    validate_resource_group(resource_group_name)

    if cmd.cli_ctx.local_context.is_on:
        cmd.cli_ctx.local_context.set(['postgres flexible-server'], 'server_name',
                                      server_name)  # Setting the server name in the local context
        cmd.cli_ctx.local_context.set(['postgres flexible-server'], 'administrator_login',
                                      user)  # Setting the server name in the local context
        cmd.cli_ctx.local_context.set(['postgres flexible-server'], 'database_name',
                                      POSTGRES_DB_NAME)  # Setting the server name in the local context
        cmd.cli_ctx.local_context.set([ALL], 'location',
                                      location)  # Setting the location in the local context
        cmd.cli_ctx.local_context.set([ALL], 'resource_group_name', resource_group_name)


def _form_response(username, sku, location, server_id, host, version, password, firewall_id=None,
                   subnet_id=None, is_password_auth=True, is_microsoft_entra_auth_enabled=False, microsoft_admin=None):

    connection_kwargs = {
        'user': username if username is not None else '{user}',
        'host': host,
        'password': password if password is not None else '{password}',
        'database': POSTGRES_DB_NAME,
    }
    output = {
        'host': host,
        'username': username if is_password_auth else None,
        'password': password if is_password_auth else None,
        'skuname': sku,
        'location': location,
        'id': server_id,
        'version': version,
        'databaseName': POSTGRES_DB_NAME,
        'connectionString': 'postgresql://{user}:{password}@{host}/{database}?sslmode=require'.format(**connection_kwargs)
    }
    if is_microsoft_entra_auth_enabled:
        user = quote(microsoft_admin) if microsoft_admin else '<admin>'
        connection_kwargs = {
            'user': user,
            'host': host,
            'database': POSTGRES_DB_NAME,
        }
        output['admin'] = microsoft_admin
        output['connectionStringMicrosoftEntra'] = 'postgresql://{user}:<access-token>@{host}/{database}?sslmode=require'.format(**connection_kwargs)
    if firewall_id is not None:
        output['firewallName'] = firewall_id
    if subnet_id is not None:
        output['subnetId'] = subnet_id
    return output


def flexible_server_restore(cmd, client,
                            resource_group_name, server_name,
                            source_server, restore_point_in_time=None, zone=None, no_wait=False,
                            subnet=None, vnet=None,
                            private_dns_zone_arguments=None, geo_redundant_backup=None,
                            byok_identity=None, byok_key=None, backup_byok_identity=None, backup_byok_key=None,
                            federated_client_id=None, backup_federated_client_id=None,
                            storage_type=None, yes=False):

    server_name = server_name.lower()

    validate_resource_group(resource_group_name)

    if not is_valid_resource_id(source_server):
        if len(source_server.split('/')) == 1:
            source_server_id = resource_id(
                subscription=get_subscription_id(cmd.cli_ctx),
                resource_group=resource_group_name,
                namespace=RESOURCE_PROVIDER,
                type='flexibleServers',
                name=source_server)
        else:
            raise ValueError('The provided source server {} is invalid.'.format(source_server))
    else:
        source_server_id = source_server

    restore_point_in_time = validate_and_format_restore_point_in_time(restore_point_in_time)

    try:
        id_parts = parse_resource_id(source_server_id)
        source_subscription_id = id_parts['subscription']
        postgres_source_client = get_postgresql_flexible_management_client(cmd.cli_ctx, source_subscription_id)
        source_server_object = postgres_source_client.servers.get(id_parts['resource_group'], id_parts['name'])

        location = ''.join(source_server_object.location.lower().split())

        db_context = DbContext(
            cmd=cmd, azure_sdk=postgresql_flexibleservers, cf_firewall=cf_postgres_flexible_firewall_rules,
            cf_db=cf_postgres_flexible_db, cf_availability=cf_postgres_check_resource_availability,
            cf_private_dns_zone_suffix=cf_postgres_flexible_private_dns_zone_suffix_operations,
            logging_name='PostgreSQL', command_group='postgres', server_client=client, location=location)
        validate_server_name(db_context, server_name, 'Microsoft.DBforPostgreSQL/flexibleServers')

        pg_byok_validator(byok_identity, byok_key, backup_byok_identity, backup_byok_key, geo_redundant_backup,
                          federated_client_id=federated_client_id,
                          backup_federated_client_id=backup_federated_client_id)

        pg_restore_validator(source_server_object.sku.tier, storage_type=storage_type)
        storage = postgresql_flexibleservers.models.Storage(type=storage_type if source_server_object.storage.type != "PremiumV2_LRS" else None)

        parameters = postgresql_flexibleservers.models.Server(
            location=location,
            point_in_time_utc=restore_point_in_time,
            source_server_resource_id=source_server_id,  # this should be the source server name, not id
            create_mode="PointInTimeRestore",
            availability_zone=zone,
            storage=storage
        )

        if source_server_object.network.public_network_access == 'Disabled' and any((vnet, subnet)):
            parameters.network, _, _ = build_network_configuration(
                cmd=cmd,
                resource_group_name=resource_group_name,
                server_name=server_name,
                location=location,
                db_context=db_context,
                private_dns_zone_arguments=private_dns_zone_arguments,
                public_access='Disabled',
                vnet=vnet,
                subnet=subnet,
                yes=yes)
        else:
            parameters.network = source_server_object.network

        parameters.backup = postgresql_flexibleservers.models.Backup(geo_redundant_backup=geo_redundant_backup)

        parameters.identity, parameters.data_encryption = build_identity_and_data_encryption(byok_identity=byok_identity,
                                                                                             byok_key=byok_key,
                                                                                             backup_byok_identity=backup_byok_identity,
                                                                                             backup_byok_key=backup_byok_key,
                                                                                             federated_client_id=federated_client_id,
                                                                                             backup_federated_client_id=backup_federated_client_id)

    except Exception as e:
        raise ResourceNotFoundError(e)

    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, server_name, parameters)


# pylint: disable=too-many-branches
def flexible_server_update_custom_func(cmd, client, instance,
                                       sku_name=None, tier=None,
                                       storage_gb=None,
                                       backup_retention=None,
                                       administrator_login_password=None,
                                       zonal_resiliency=None,
                                       allow_same_zone=False,
                                       standby_availability_zone=None,
                                       maintenance_window=None,
                                       byok_identity=None, byok_key=None,
                                       backup_byok_identity=None, backup_byok_key=None,
                                       federated_client_id=None, backup_federated_client_id=None,
                                       microsoft_entra_auth=None, password_auth=None,
                                       private_dns_zone_arguments=None,
                                       public_access=None,
                                       tags=None,
                                       auto_grow=None,
                                       performance_tier=None,
                                       iops=None, throughput=None,
                                       cluster_size=None, yes=False):

    # validator
    location = ''.join(instance.location.lower().split())
    db_context = DbContext(
        cmd=cmd, azure_sdk=postgresql_flexibleservers, cf_firewall=cf_postgres_flexible_firewall_rules,
        cf_db=cf_postgres_flexible_db, cf_availability=cf_postgres_check_resource_availability,
        cf_private_dns_zone_suffix=cf_postgres_flexible_private_dns_zone_suffix_operations,
        logging_name='PostgreSQL', command_group='postgres', server_client=client, location=location)

    pg_arguments_validator(db_context,
                           location=location,
                           tier=tier,
                           sku_name=sku_name,
                           storage_gb=storage_gb,
                           auto_grow=auto_grow,
                           iops=iops,
                           throughput=throughput,
                           zonal_resiliency=zonal_resiliency,
                           allow_same_zone=allow_same_zone,
                           zone=instance.availability_zone,
                           standby_availability_zone=standby_availability_zone,
                           byok_identity=byok_identity,
                           byok_key=byok_key,
                           backup_byok_identity=backup_byok_identity,
                           backup_byok_key=backup_byok_key,
                           federated_client_id=federated_client_id,
                           backup_federated_client_id=backup_federated_client_id,
                           performance_tier=performance_tier,
                           cluster_size=cluster_size, instance=instance)

    server_module_path = instance.__module__
    module = import_module(server_module_path)
    ServerForPatch = getattr(module, 'ServerForPatch')

    server_id_parts = parse_resource_id(instance.id)
    resource_group_name = server_id_parts['resource_group']
    server_name = server_id_parts['name']

    if public_access:
        instance.network.public_network_access = public_access

    if private_dns_zone_arguments:
        private_dns_zone_id = resolve_private_dns_zone_id(
            db_context,
            resource_group_name,
            server_name,
            private_dns_zone=private_dns_zone_arguments,
            subnet_id=instance.network.delegated_subnet_resource_id,
            location=location)
        instance.network.private_dns_zone_arm_resource_id = private_dns_zone_id

    _confirm_restart_server(instance, sku_name, storage_gb, yes)

    if sku_name:
        instance.sku.name = sku_name

    if tier:
        instance.sku.tier = tier

    if storage_gb:
        instance.storage.storage_size_gb = storage_gb

    if auto_grow:
        instance.storage.auto_grow = auto_grow

    instance.storage.tier = performance_tier if performance_tier else None

    if instance.storage.type == "PremiumV2_LRS":
        instance.storage.tier = None

        if iops:
            instance.storage.iops = iops

        if throughput:
            instance.storage.throughput = throughput
    else:
        instance.storage.type = None
        instance.storage.iops = None
        instance.storage.throughput = None

    if backup_retention:
        instance.backup.backup_retention_days = backup_retention

    if maintenance_window:
        if maintenance_window.lower() == "disabled":
            # if disabled is pass in reset to default values
            day_of_week = start_hour = start_minute = 0
            custom_window = "Disabled"
        else:
            day_of_week, start_hour, start_minute = parse_maintenance_window(maintenance_window)
            custom_window = "Enabled"

        # set values - if maintenance_window when is None when created then create a new object
        instance.maintenance_window.day_of_week = day_of_week
        instance.maintenance_window.start_hour = start_hour
        instance.maintenance_window.start_minute = start_minute
        instance.maintenance_window.custom_window = custom_window

    identity, data_encryption = build_identity_and_data_encryption(byok_identity=byok_identity,
                                                                   byok_key=byok_key,
                                                                   backup_byok_identity=backup_byok_identity,
                                                                   backup_byok_key=backup_byok_key,
                                                                   federated_client_id=federated_client_id,
                                                                   backup_federated_client_id=backup_federated_client_id,
                                                                   instance=instance)

    auth_config = instance.auth_config
    administrator_login = instance.administrator_login if instance.administrator_login else None
    if microsoft_entra_auth:
        auth_config.active_directory_auth = microsoft_entra_auth
    if password_auth:
        administrator_login, administrator_login_password = _update_login(server_name, resource_group_name, auth_config,
                                                                          password_auth, administrator_login, administrator_login_password)
        auth_config.password_auth = password_auth

    if cluster_size:
        instance.cluster.cluster_size = cluster_size

    params = ServerForPatch(sku=instance.sku,
                            storage=instance.storage,
                            backup=instance.backup,
                            administrator_login=administrator_login,
                            administrator_login_password=administrator_login_password,
                            availability_zone=instance.availability_zone,
                            maintenance_window=instance.maintenance_window,
                            network=instance.network,
                            identity=identity,
                            data_encryption=data_encryption,
                            auth_config=auth_config,
                            cluster=instance.cluster,
                            tags=tags)

    # High availability can't be updated with existing properties
    high_availability_param = postgresql_flexibleservers.models.HighAvailability()
    high_availability = None
    if zonal_resiliency is not None:
        if zonal_resiliency.lower() == 'disabled':
            high_availability = 'Disabled'
        else:
            list_capability_info = get_postgres_server_capability_info(cmd, resource_group_name, server_name)
            single_az = list_capability_info['single_az']
            high_availability = 'SameZone' if single_az and allow_same_zone else 'ZoneRedundant'
    if high_availability:
        high_availability_param.mode = high_availability

        if high_availability.lower() != "disabled" and standby_availability_zone:
            high_availability_param.standby_availability_zone = standby_availability_zone

        # PG 11 and 12 will never receive fabric mirroring support. Ignite 2025 Fabric mirroring supported on 17. Skip this check for servers of these versions
        if high_availability.lower() != "disabled" and str(instance.version) not in ["11", "12", "17", "18"]:
            config_client = cf_postgres_flexible_config(cmd.cli_ctx, '_')
            fabric_mirror_status = config_client.get(resource_group_name, server_name, 'azure.fabric_mirror_enabled')
            if (fabric_mirror_status and fabric_mirror_status.value.lower() == 'on'):
                raise CLIError("High availability cannot be enabled while Fabric mirroring is Active. Please disable Fabric mirroring to enable high availability.")

        params.high_availability = high_availability_param

    return params


# Common functions used by other providers
def flexible_server_update_get(client, resource_group_name, server_name):
    validate_resource_group(resource_group_name)

    return client.get(resource_group_name, server_name)


def flexible_server_update_set(client, resource_group_name, server_name, parameters):
    validate_resource_group(resource_group_name)

    return client.begin_update(resource_group_name, server_name, parameters)


def _update_login(server_name, resource_group_name, auth_config, password_auth, administrator_login, administrator_login_password):
    if auth_config.password_auth.lower() == 'disabled' and password_auth.lower() == 'enabled':
        administrator_login = administrator_login if administrator_login else prompt('Please enter administrator username for the server. Once set, it cannot be changed: ')
        if not administrator_login:
            raise CLIError('Administrator username is required for enabling password authentication.')
        if not administrator_login_password:
            administrator_login_password = generate_password(administrator_login_password)
            logger.warning('Make a note of password "%s". You can '
                           'reset your password with "az postgres flexible-server update -n %s -g %s -p <new-password>".',
                           administrator_login_password, server_name, resource_group_name)

    return administrator_login, administrator_login_password


# pylint: disable=chained-comparison
def _confirm_restart_server(instance, sku_name, storage_gb, yes):
    show_confirmation = False

    # check if sku_name is changed
    if sku_name and sku_name != instance.sku.name:
        show_confirmation = True

    # check if requested storage growth is crossing the 4096 threshold
    if storage_gb and storage_gb > 4096 and instance.storage.storage_size_gb <= 4096 and instance.storage.type != "PremiumV2_LRS":
        show_confirmation = True

    if not yes and show_confirmation:
        user_confirmation('You are trying to update the compute or storage size assigned to your server in a way that '
                          'requires a server restart. During the restart, you\'ll experience some downtime of the server. Do you want to proceed?', yes=yes)


def flexible_server_delete(cmd, client, resource_group_name, server_name, yes=False):
    validate_resource_group(resource_group_name)
    result = None
    if not yes:
        user_confirmation(
            "Are you sure you want to delete the server '{0}' in resource group '{1}'".format(server_name,
                                                                                              resource_group_name), yes=yes)
    try:
        result = client.begin_delete(resource_group_name, server_name)
        if cmd.cli_ctx.local_context.is_on:
            local_context_file = cmd.cli_ctx.local_context._get_local_context_file()  # pylint: disable=protected-access
            local_context_file.remove_option('postgres flexible-server', 'server_name')
            local_context_file.remove_option('postgres flexible-server', 'administrator_login')
            local_context_file.remove_option('postgres flexible-server', 'database_name')
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(ex)
        raise CLIError(ex)
    return result


def flexible_server_postgresql_get(cmd, resource_group_name, server_name):
    validate_resource_group(resource_group_name)
    client = get_postgresql_flexible_management_client(cmd.cli_ctx)
    return client.servers.get(resource_group_name, server_name)


def server_list_custom_func(client, resource_group_name=None, show_cluster=None):
    if not check_resource_group(resource_group_name):
        resource_group_name = None

    servers = client.list_by_subscription()

    if resource_group_name:
        servers = client.list_by_resource_group(resource_group_name)

    if show_cluster:
        servers = [s for s in servers if s.cluster is not None]
    else:
        servers = [s for s in servers if s.cluster is None]

    return servers


def flexible_list_skus(cmd, client, location):
    result = client.list(location)
    logger.warning('For prices please refer to https://aka.ms/postgres-pricing')
    return result


def flexible_server_georestore(cmd, client, resource_group_name, server_name, source_server, location, zone=None,
                               vnet=None, subnet=None,
                               private_dns_zone_arguments=None, geo_redundant_backup=None, no_wait=False, yes=False,
                               byok_identity=None, byok_key=None, backup_byok_identity=None, backup_byok_key=None,
                               federated_client_id=None, backup_federated_client_id=None,
                               restore_point_in_time=None):
    validate_resource_group(resource_group_name)

    server_name = server_name.lower()

    if not is_valid_resource_id(source_server):
        if _is_resource_name(source_server):
            source_server_id = resource_id(subscription=get_subscription_id(cmd.cli_ctx),
                                           resource_group=resource_group_name,
                                           namespace='Microsoft.DBforPostgreSQL',
                                           type='flexibleServers',
                                           name=source_server)
        else:
            raise CLIError('The provided source-server {} is invalid.'.format(source_server))
    else:
        source_server_id = source_server

    restore_point_in_time = validate_and_format_restore_point_in_time(restore_point_in_time)

    try:
        id_parts = parse_resource_id(source_server_id)
        validate_citus_cluster(cmd, id_parts['resource_group'], id_parts['name'])
        source_subscription_id = id_parts['subscription']
        postgres_source_client = get_postgresql_flexible_management_client(cmd.cli_ctx, source_subscription_id)
        source_server_object = postgres_source_client.servers.get(id_parts['resource_group'], id_parts['name'])
    except Exception as e:
        raise ResourceNotFoundError(e)

    db_context = DbContext(
        cmd=cmd, azure_sdk=postgresql_flexibleservers, cf_firewall=cf_postgres_flexible_firewall_rules,
        cf_db=cf_postgres_flexible_db, cf_availability=cf_postgres_check_resource_availability,
        cf_private_dns_zone_suffix=cf_postgres_flexible_private_dns_zone_suffix_operations,
        logging_name='PostgreSQL', command_group='postgres', server_client=client, location=location)

    validate_server_name(db_context, server_name, 'Microsoft.DBforPostgreSQL/flexibleServers')
    if source_server_object.network.delegated_subnet_resource_id is not None:
        validate_georestore_network(source_server_object, None, vnet, subnet, 'postgres')

    pg_byok_validator(byok_identity, byok_key, backup_byok_identity, backup_byok_key, geo_redundant_backup,
                      federated_client_id=federated_client_id,
                      backup_federated_client_id=backup_federated_client_id)

    storage = postgresql_flexibleservers.models.Storage(type=None)

    parameters = postgresql_flexibleservers.models.Server(
        point_in_time_utc=restore_point_in_time,
        location=location,
        source_server_resource_id=source_server_id,
        create_mode="GeoRestore",
        availability_zone=zone,
        storage=storage
    )

    if source_server_object.network.public_network_access == 'Disabled':
        parameters.network, _, _ = build_network_configuration(
            cmd=cmd,
            resource_group_name=resource_group_name,
            server_name=server_name,
            location=location,
            db_context=db_context,
            private_dns_zone_arguments=private_dns_zone_arguments,
            public_access='Disabled',
            vnet=vnet,
            subnet=subnet,
            yes=yes)

    parameters.backup = postgresql_flexibleservers.models.Backup(geo_redundant_backup=geo_redundant_backup)

    parameters.identity, parameters.data_encryption = build_identity_and_data_encryption(byok_identity=byok_identity,
                                                                                         byok_key=byok_key,
                                                                                         backup_byok_identity=backup_byok_identity,
                                                                                         backup_byok_key=backup_byok_key,
                                                                                         federated_client_id=federated_client_id,
                                                                                         backup_federated_client_id=backup_federated_client_id)

    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, server_name, parameters)


def flexible_server_revivedropped(cmd, client, resource_group_name, server_name, source_server, location, zone=None,
                                  vnet=None, subnet=None,
                                  private_dns_zone_arguments=None, geo_redundant_backup=None, no_wait=False, yes=False,
                                  byok_identity=None, byok_key=None, backup_byok_identity=None, backup_byok_key=None):
    validate_resource_group(resource_group_name)

    server_name = server_name.lower()

    if not is_valid_resource_id(source_server):
        if _is_resource_name(source_server):
            source_server_id = resource_id(subscription=get_subscription_id(cmd.cli_ctx),
                                           resource_group=resource_group_name,
                                           namespace='Microsoft.DBforPostgreSQL',
                                           type='flexibleServers',
                                           name=source_server)
        else:
            raise CLIError('The provided source-server {} is invalid.'.format(source_server))
    else:
        source_server_id = source_server

    db_context = DbContext(
        cmd=cmd, azure_sdk=postgresql_flexibleservers, cf_firewall=cf_postgres_flexible_firewall_rules,
        cf_db=cf_postgres_flexible_db, cf_availability=cf_postgres_check_resource_availability,
        cf_private_dns_zone_suffix=cf_postgres_flexible_private_dns_zone_suffix_operations,
        logging_name='PostgreSQL', command_group='postgres', server_client=client, location=location)

    validate_server_name(db_context, server_name, 'Microsoft.DBforPostgreSQL/flexibleServers')

    pg_byok_validator(byok_identity, byok_key, backup_byok_identity, backup_byok_key, geo_redundant_backup)

    storage = postgresql_flexibleservers.models.Storage(type=None)

    parameters = postgresql_flexibleservers.models.Server(
        point_in_time_utc=get_current_time(),
        location=location,
        source_server_resource_id=source_server_id,
        create_mode="ReviveDropped",
        availability_zone=zone,
        storage=storage
    )

    if vnet is not None or subnet is not None or private_dns_zone_arguments is not None:
        parameters.network, _, _ = build_network_configuration(
            cmd=cmd,
            resource_group_name=resource_group_name,
            server_name=server_name,
            location=location,
            db_context=db_context,
            private_dns_zone_arguments=private_dns_zone_arguments,
            public_access='Disabled',
            vnet=vnet,
            subnet=subnet,
            yes=yes)

    parameters.backup = postgresql_flexibleservers.models.Backup(geo_redundant_backup=geo_redundant_backup)

    parameters.identity, parameters.data_encryption = build_identity_and_data_encryption(byok_identity=byok_identity,
                                                                                         byok_key=byok_key,
                                                                                         backup_byok_identity=backup_byok_identity,
                                                                                         backup_byok_key=backup_byok_key)

    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, server_name, parameters)


def flexible_server_connection_string(
        server_name='{server}',
        database_name='{database}',
        administrator_login='{login}',
        administrator_login_password='{password}',
        show_pg_bouncer=False):
    host = '{}.postgres.database.azure.com'.format(server_name)
    port = 5432
    if show_pg_bouncer is True:
        port = 6432

    return {
        'connectionStrings': _create_postgresql_connection_strings(host, administrator_login,
                                                                   administrator_login_password, database_name, port)
    }


def _create_postgresql_connection_strings(host, user, password, database, port):

    result = {
        'psql_cmd': "postgresql://{user}:{password}@{host}/{database}?sslmode=require",
        'ado.net': "Server={host};Database={database};Port={port};User Id={user};Password={password};Ssl Mode=Require;",
        'jdbc': "jdbc:postgresql://{host}:{port}/{database}?user={user}&password={password}&sslmode=require",
        'jdbc Spring': "spring.datasource.url=jdbc:postgresql://{host}:{port}/{database}  "
                       "spring.datasource.username={user}  "
                       "spring.datasource.password={password}",
        'node.js': "var conn= new Client({open_brace}host:'{host}', user:'{user}', password:'{password}', database:'{database}', port:{port}, ssl:{open_brace}ca:fs.readFileSync(\"{ca-cert filename}\"){close_brace}{close_brace});",
        'php': "pg_connect(\"host={host} port={port} dbname={database} user={user} password={password}\");",
        'python': "cnx = psycopg2.connect(user='{user}', password='{password}', host='{host}', "
                  "port={port}, database='{database}')",
        'ruby': "connection = PG::Connection.new(user => \"{user}\", password => \"{password}\", database => \"{database}\", host => \"{host}\", "
                "port => '{port}')",
    }

    connection_kwargs = {
        'host': host,
        'user': user,
        'password': password if password is not None else '{password}',
        'database': database,
        'port': port,
        'open_brace': '{',
        'close_brace': '}',
        'ca-cert filename': '{ca-cert filename}'
    }

    for k, v in result.items():
        result[k] = v.format(**connection_kwargs)
    return result
